from babelfont.convertors import Convert
import warnings


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return "# [warning] %s\n" % (message)


warnings.formatwarning = warning_on_one_line

f = Convert("Nunito3.glyphs").load()
f.save("output/test.babelfont")
f.save("output/test.ttf")




import logging
from typing import Dict, List
import uuid

import ufoLib2
from fontTools import designspaceLib
from fontTools.designspaceLib import DesignSpaceDocument

from babelfont import (
    Features,
    Font,
    Master,
    Instance,
    Glyph,
    Layer,
    Shape,
    Node,
    Anchor,
    Guide,
    Axis,
    I18NDictionary,
)
from babelfont.BaseObject import OTValue
from babelfont.convertors import BaseConvertor

log = logging.getLogger(__name__)

UFO_KEY = "org.unifiedfontobject"

metrics = {
    "xHeight": "xHeight",
    "capHeight": "capHeight",
    "ascender": "ascender",
    "descender": "descender",
    "italicAngle": "italicAngle",
    "hheaAscender": "openTypeHheaAscender",
    "hheaDescender": "openTypeHheaDescender",
    "hheaLineGap": "openTypeHheaLineGap",
    "winAscent": "openTypeOS2WinAscent",
    "winDescent": "openTypeOS2WinDescent",
    "typoAscender": "openTypeOS2TypoAscender",
    "typoDescender": "openTypeOS2TypoDescender",
    "typoLineGap": "openTypeOS2TypoLineGap",
    "subscriptXSize": "openTypeOS2SubscriptXSize",
    "subscriptYSize": "openTypeOS2SubscriptYSize",
    "subscriptXOffset": "openTypeOS2SubscriptXOffset",
    "subscriptYOffset": "openTypeOS2SubscriptYOffset",
    "superscriptXSize": "openTypeOS2SuperscriptXSize",
    "superscriptYSize": "openTypeOS2SuperscriptYSize",
    "superscriptXOffset": "openTypeOS2SuperscriptXOffset",
    "superscriptYOffset": "openTypeOS2SuperscriptYOffset",
    "strikeoutSize": "openTypeOS2StrikeoutSize",
    "strikeoutPosition": "openTypeOS2StrikeoutPosition",
    "underlinePosition": "postscriptUnderlinePosition",
    "underlineThickness": "postscriptUnderlineThickness",
    "hheaCaretSlopeRise": "openTypeHheaCaretSlopeRise",
    "hheaCaretSlopeRun": "openTypeHheaCaretSlopeRun",
    "hheaCaretOffset": "openTypeHheaCaretOffset",
}

# Unicode ranges / code page ranges are special

custom_opentype_values = {
    "openTypeGaspRangeRecords": ("GASP", "gaspRange"),
    # XX more gasp here
    "openTypeHeadCreated": ("head", "created"),
    "openTypeHeadLowestRecPPEM": ("head", "lowestRecPPEM"),
    "openTypeHeadFlags": ("head", "flags"),
    "openTypeOS2WidthClass": ("OS/2", "usWidthClass"),
    "openTypeOS2WeightClass": ("OS/2", "usWeightClass"),
    "openTypeOS2Selection": ("OS/2", "fsSelection"),
    "openTypeOS2VendorID": ("OS/2", "achVendID"),
    "openTypeOS2Panose": ("OS/2", "bPanose"),
    "openTypeOS2FamilyClass": ("OS/2", "sFamilyClass"),
    "openTypeOS2Type": ("OS/2", "fsType"),
}

BITARRAY = [
    ("head", "flags"),
    ("GASP", "gaspRange"),
    ("OS/2", "fsType"),
    ("OS/2", "fsSelection"),
]


def bitarray_to_int(value):
    return sum(2**i for i in value)


def int_to_bitarray(value):
    return [i for i in range(32) if value & 2**i]


def ribbi(style: str) -> bool:
    return style.lower() in ["regular", "italic", "bold", "bold italic"]


class Designspace(BaseConvertor):
    suffix = ".designspace"

    @classmethod
    def load(cls, convertor, compile_only=False):
        self = cls()
        self.ds = DesignSpaceDocument.fromfile(convertor.filename)
        self.ds.loadSourceFonts(ufoLib2.Font.open)
        self.font = Font()
        return self._load()

    def _load(self):
        self._load_axes()

        for source in self.ds.sources:
            source._babelfont_master = self._load_master(source)
            self.font.masters.append(source._babelfont_master)

        for instance in self.ds.instances:
            self.font.instances.append(self._load_instance(instance))

        firstmaster = self.ds.sources[0].font
        self._load_metadata(firstmaster)
        glyphs_dict = self._load_glyphs(firstmaster)

        # Right, let's find all the layers. This will be messy.
        for source in self.ds.sources:
            for ufo_layer in source.font.layers:
                for g in source.font.keys():
                    if g not in glyphs_dict:
                        log.warning(
                            "Incompatible glyph set: %s appears in %s but is not in default",
                            g,
                            source.filename,
                        )
                        continue
                    if g not in ufo_layer:
                        continue
                    glyphs_dict[g].layers.append(self._load_layer(source, ufo_layer, g))
        self._fixup_glyph_exported(self.ds.sources[0].font)
        return self.font

    def _fixup_glyph_exported(self, ufo: ufoLib2.Font):
        for glyph in ufo.lib.get("public.skipExportGlyphs", []):
            self.font.glyphs[glyph].exported = False

    def _load_glyphs(self, master: ufoLib2.Font):
        glyphs_dict = {}
        # Start with glyph order if there is one
        order = master.lib.get("public.glyphOrder", [])
        # Then add any remaining glyphs
        order += [g for g in master.keys() if g not in order]
        for g in order:
            glyphs_dict[g] = self._load_glyph(master[g])
            self.font.glyphs.append(glyphs_dict[g])
        return glyphs_dict

    def _load_axes(self):
        for a in self.ds.axes:
            self.font.axes.append(
                Axis(
                    name=a.name,
                    tag=a.tag,
                    min=a.minimum,
                    max=a.maximum,
                    default=a.default,
                    map=a.map,
                )
            )

    def _load_master(self, source: designspaceLib.SourceDescriptor):
        font: ufoLib2.Font = source.font
        i = font.info
        master = Master(
            name=source.name,
            id=(source.name or uuid.uuid1()),
        )
        for our_metric, their_metric in metrics.items():
            master.metrics[our_metric] = getattr(i, their_metric)
        _axis_name_to_id = {a.name.get_default(): a.tag for a in self.font.axes}
        # names XXX
        # guidelines
        master.guides = [self._load_guide(g) for g in (i.guidelines or [])]
        master.location = {_axis_name_to_id[k]: v for k, v in source.location.items()}
        master.font = self.font
        master.kerning = self._load_kerning(source)
        for key, value in font.lib.items():
            if not key.startswith("public."):
                if UFO_KEY not in master._formatspecific:
                    master._formatspecific[UFO_KEY] = {}
                master._formatspecific[UFO_KEY][key] = value

        master.features = Features.from_fea(font.features.text)
        self._load_groups(font.groups)
        assert master.valid
        return master

    def _load_groups(self, groups: Dict[str, List[str]]):
        for name, value in groups.items():
            self.font.features.classes[name] = value

    def _load_guide(self, ufo_guide: ufoLib2.objects.Guideline):
        return Guide(
            pos=[ufo_guide.x, ufo_guide.y, ufo_guide.angle],
            name=ufo_guide.name,
            color=ufo_guide.color,
        )

    def _load_kerning(self, source: designspaceLib.SourceDescriptor):
        font: ufoLib2.Font = source.font
        kerning = {}
        for (left, right), value in font.kerning.items():
            if left.startswith("public.kern"):
                left = "@" + left
            if right.startswith("public.kern"):
                right = "@" + right
            kerning[(left, right)] = value
        return kerning

    def _load_glyph(self, ufo_glyph: ufoLib2.objects.Glyph):
        cp = []
        if ufo_glyph.unicodes or ufo_glyph.unicode:
            cp = ufo_glyph.unicodes or [ufo_glyph.unicode]
        lib = self.ds.sources[0].font.lib
        category = lib.get("public.openTypeCategories", {}).get(ufo_glyph.name, "base")
        g = Glyph(name=ufo_glyph.name, codepoints=cp, category=category)

        if "public.postscriptNames" in lib and g.name in lib["public.postscriptNames"]:
            g.production_name = lib["public.postscriptNames"][g.name]
        return g

    def _load_layer(
        self, source: designspaceLib.SourceDescriptor, ufo_layer, glyphname: str
    ):
        if ufo_layer.name == "public.default":
            layer_id = source._babelfont_master.id
        else:
            layer_id = ufo_layer.name
        ufo_glyph = ufo_layer[glyphname]
        width = ufo_glyph.width
        layer = Layer(width=width, id=layer_id)
        layer._master = source._babelfont_master.id
        layer._font = self.font
        layer._glyph = self.font.glyphs[glyphname]
        for contour in ufo_glyph:
            layer.shapes.append(self._load_contour(contour))
        for component in ufo_glyph.components:
            layer.shapes.append(self._load_component(component))
        for anchor in ufo_glyph.anchors:
            layer.anchors.append(self._load_anchor(anchor))
        assert layer.valid
        return layer

    def _load_component(self, shape: ufoLib2.objects.Component):
        c = Shape(ref=shape.baseGlyph, transform=shape.transformation)
        return c

    def _load_anchor(self, anchor: ufoLib2.objects.Anchor):
        return Anchor(name=anchor.name, x=int(anchor.x), y=int(anchor.y))

    def _load_contour(self, contour):
        shape = Shape()
        shape.nodes = []
        for p in contour:
            segtype = p.segmentType
            if p.segmentType == "move":
                segtype = "line"
            ourtype = Node._from_pen_type[segtype]
            if p.smooth:
                ourtype = ourtype + "s"
            shape.nodes.append(Node(p.x, p.y, ourtype))
        return shape

    def _load_instance(self, ufo_instance):
        _axis_tag = {axis.name.get_default(): axis.tag for axis in self.font.axes}
        location = {_axis_tag[k]: v for k, v in ufo_instance.location.items()}
        instance = Instance(name=ufo_instance.name, location=location)
        return instance

    names_dict = {
        "designer": "openTypeNameDesigner",
        "designerURL": "openTypeNameDesignerURL",
        "manufacturer": "openTypeNameManufacturer",
        "manufacturerURL": "openTypeNameManufacturerURL",
        "license": "openTypeNameLicense",
        "licenseURL": "openTypeNameLicenseURL",
        "version": "openTypeNameVersion",
        "uniqueID": "openTypeNameUniqueID",
        "description": "openTypeNameDescription",
        "compatibleFullName": "openTypeNameCompatibleFullName",
        "sampleText": "openTypeNameSampleText",
        "WWSFamilyName": "openTypeNameWWSFamilyName",
        "WWSSubfamilyName": "openTypeNameWWSSubfamilyName",
        "copyright": "copyright",
        "styleMapFamilyName": "styleMapFamilyName",
        "familyName": "familyName",
        "trademark": "trademark",
        "styleName": "styleName",
        "styleMapStyleName": "styleMapStyleName",
        "preferredSubfamilyName": "openTypeNamePreferredSubfamilyName",
    }

    def _load_metadata(self, ufo):
        firstfontinfo = ufo.info
        self.font.upm = firstfontinfo.unitsPerEm
        self.font.version = (firstfontinfo.versionMajor, firstfontinfo.versionMinor)
        self.font.note = firstfontinfo.note
        for ours, theirs in self.names_dict.items():
            their_value = getattr(firstfontinfo, theirs)
            if their_value:
                getattr(self.font.names, ours).set_default(their_value)
        for ufofield, (table, field) in custom_opentype_values.items():
            if getattr(firstfontinfo, ufofield) is not None:
                value = getattr(firstfontinfo, ufofield)
                if (table, field) in BITARRAY:
                    value = bitarray_to_int(value)
                self.font.customOpenTypeValues.append(
                    OTValue(table=table, field=field, value=value)
                )

    def _save(self):
        self.ds = DesignSpaceDocument()
        self.save_axes()
        self.save_sources()
        self.save_instances()
        # Lib
        unexported = [g.name for g in self.font.glyphs if not g.exported]
        if unexported:
            self.ds.lib["public.skipExportGlyphs"] = sorted(unexported)
        self.ds.write(self.filename)

    def save_axes(self):
        axes = self.font.axes
        if not axes:
            axes = [
                Axis(
                    name=I18NDictionary.with_default("Weight"),
                    tag="wght",
                    min=100,
                    max=100,
                    default=100,
                )
            ]
        for axis in axes:
            axisDescriptor = designspaceLib.AxisDescriptor()
            axisDescriptor.name = axis.name.get_default()
            other_names = axis.name.as_fonttools_dict
            if other_names:
                axis.labelNames = other_names
            axisDescriptor.tag = axis.tag
            axisDescriptor.minimum = axis.min
            axisDescriptor.maximum = axis.max
            axisDescriptor.default = axis.default
            axisDescriptor.map = axis.map
            self.ds.addAxis(axisDescriptor)

    def _master_filename(self, master):
        return (
            self.font.names.familyName.get_default().replace(" ", "")
            + "-"
            + master.name.get_default().replace(" ", "")
            + ".ufo"
        )

    def save_sources(self):
        font = self.font
        for master in font.masters:
            sourceDescriptor = designspaceLib.SourceDescriptor()
            sourceDescriptor.name = (
                font.names.familyName.get_default() + " " + master.name.get_default()
            )
            sourceDescriptor.filename = self._master_filename(master)
            sourceDescriptor.styleName = master.name.get_default()
            sourceDescriptor.familyName = font.names.familyName.get_default()
            if master == self.font.default_master:
                sourceDescriptor.copyLib = True
                sourceDescriptor.copyGroups = True
                sourceDescriptor.copyFeatures = True
                sourceDescriptor.copyInfo = True
            # sourceDescriptor.filename = master.filename
            # sourceDescriptor.path = master.filename
            axis_tags_to_names = {
                axis.tag: axis.name.get_default() for axis in self.font.axes
            }
            if master.location:
                sourceDescriptor.location = {
                    axis_tags_to_names[ax]: value
                    for ax, value in master.location.items()
                }
            else:
                sourceDescriptor.location = {"Weight": 100}
            self.ds.addSource(sourceDescriptor)
            self.save_master_to_ufo(
                master,
                self._master_filename(master),
                is_default=(master == self.font.default_master),
            )

    def save_instances(self):
        for instance in self.font.instances:
            instanceDescriptor = designspaceLib.InstanceDescriptor()
            instanceDescriptor.name = (
                self.font.names.familyName.get_default()
                + " "
                + instance.name.get_default()
            )
            names = self.font.names
            instanceDescriptor.familyName = names.familyName.get_default()
            instanceDescriptor.styleName = instance.name.get_default()
            instanceDescriptor.styleMapFamilyName = (
                names.styleMapFamilyName.get_default() or names.familyName.get_default()
            )
            if not ribbi(instance.name.get_default()):
                instanceDescriptor.styleMapFamilyName += (
                    " " + instance.name.get_default()
                )
                instanceDescriptor.styleMapStyleName = "regular"
            axis_tags_to_names = {
                axis.tag: axis.name.get_default() for axis in self.font.axes
            }
            if instance.location:
                instanceDescriptor.location = {
                    axis_tags_to_names[ax]: value
                    for ax, value in instance.location.items()
                }
            else:
                instanceDescriptor.location = {"Weight": 400}
            self.ds.addInstance(instanceDescriptor)

    def save_master_to_ufo(self, master: Master, filename, is_default=False):
        ufo = ufoLib2.Font()
        ufo.info.unitsPerEm = self.font.upm
        ufo.info.versionMajor, ufo.info.versionMinor = self.font.version
        if self.font.note and is_default:
            ufo.info.note = self.font.note
        for ours, theirs in self.names_dict.items():
            our_value = getattr(self.font.names, ours).get_default()
            if our_value is not None:
                setattr(ufo.info, theirs, getattr(self.font.names, ours).get_default())
        if not ufo.info.styleName:
            ufo.info.styleName = master.name.get_default()
        for glyph in self.font.glyphs:
            ufo.newGlyph(glyph.name)
        for glyph in self.font.glyphs:
            ufo_glyph = ufo[glyph.name]
            layer = master.get_glyph_layer(glyph.name)
            assert layer
            self.save_layer_to_ufo(ufo_glyph, layer)
            ufo_glyph.unicodes = [int(cp) for cp in glyph.codepoints]
            if layer.background is not None:
                if "public.background" not in ufo.layers:
                    ufo.layers.newLayer("public.background")
                background_layer = None
                for possible_background_layer in glyph.layers:
                    if possible_background_layer.id == layer.background:
                        background_layer = possible_background_layer
                if background_layer:
                    background_glyph = ufo.layers["public.background"].newGlyph(
                        glyph.name
                    )
                    self.save_layer_to_ufo(background_glyph, background_layer)
        # Metrics
        for our_metric, their_metric in metrics.items():
            if our_metric in master.metrics:
                metric_value = master.metrics[our_metric]
                if their_metric in ["openTypeOS2WinDescent"] and metric_value < 0:
                    metric_value = -metric_value
                setattr(ufo.info, their_metric, metric_value)
        if is_default:
            for info_tag, (table, field) in custom_opentype_values.items():
                for otv in self.font.customOpenTypeValues:
                    if otv.table == table and otv.field == field:
                        value = otv.value
                        if (table, field) in BITARRAY:
                            if info_tag == "openTypeOS2Selection":
                                # "Bits 0 (italic), 5 (bold) and 6 (regular) must not be set here"
                                value = value & 0b11111100
                            value = int_to_bitarray(value)

                        setattr(ufo.info, info_tag, value)
        # Guides
        if master.guides:
            ufo.info.guidelines = []
        for guide in master.guides:
            ufo.info.guidelines.append(
                ufoLib2.objects.Guideline(
                    x=guide.pos[0],
                    y=guide.pos[1],
                    angle=guide.pos[2] % 360,
                    name=guide.name,
                    color=guide.color,
                )
            )
        # Kerning
        for group, members in self.font.first_kern_groups.items():
            ufo.groups["public.kern1." + group] = members
        for group, members in self.font.second_kern_groups.items():
            ufo.groups["public.kern2." + group] = members
        for (left, right), value in master.kerning.items():
            if left.startswith("@first_group_"):
                left = "public.kern1." + left[13:]
            if right.startswith("@second_group_"):
                right = "public.kern2." + right[14:]
            ufo.kerning[left, right] = value

        # Features and groups
        ufo.features.text = self.font.features.to_fea()
        # Lib
        ufo.lib["public.glyphOrder"] = [g.name for g in self.font.glyphs]
        psnames = {
            g.name: g.production_name for g in self.font.glyphs if g.production_name
        }
        if psnames:
            ufo.lib["public.postscriptNames"] = psnames
        for key, value in master._formatspecific.get(UFO_KEY, {}).items():
            ufo.lib[key] = value
        self.logger.info("Saving %s", filename)
        ufo.save(filename, overwrite=True)

    def save_layer_to_ufo(self, ufo_glyph: ufoLib2.objects.Glyph, layer: Layer):
        for shape in layer.shapes:
            if shape.is_component:
                self.save_component_to_ufo(ufo_glyph, shape)
            else:
                self.save_contour_to_ufo(ufo_glyph, shape)
        for anchor in layer.anchors:
            self.save_anchor_to_ufo(ufo_glyph, anchor)
        ufo_glyph.width = layer.width
        if layer.height:
            ufo_glyph.height = layer.height

    def save_anchor_to_ufo(self, ufo_glyph: ufoLib2.objects.Glyph, anchor: Anchor):
        ufo_glyph.appendAnchor(
            ufoLib2.objects.Anchor(name=anchor.name, x=anchor.x, y=anchor.y)
        )

    def save_component_to_ufo(self, ufo_glyph: ufoLib2.objects.Glyph, shape: Shape):
        ufo_glyph.components.append(
            ufoLib2.objects.Component(
                baseGlyph=shape.ref,
                transformation=shape.transform,
            )
        )

    def save_contour_to_ufo(self, ufo_glyph: ufoLib2.objects.Glyph, shape: Shape):
        pen = ufo_glyph.getPointPen()
        pen.beginPath()
        for node in shape.nodes:
            pen.addPoint(
                (node.x, node.y),
                segmentType=Node._to_pen_type[node.type[0]],
                smooth=node.is_smooth,
            )
        pen.endPath()



from typing import Optional, Dict, Tuple
from .BaseObject import BaseObject, I18NDictionary
from dataclasses import dataclass, field
import uuid
from fontTools.varLib.models import normalizeValue


@dataclass
class _AxisFields:
    name: I18NDictionary = field(
        metadata={"description": "The display name for this axis."}
    )
    tag: str = field(metadata={"description": "The four-letter axis tag."})
    id: str = field(
        default_factory=lambda: str(uuid.uuid1()),
        repr=False,
        metadata={
            "description": """An ID used to refer to this axis in the Master,
Layer and Instance `location` fields. (This is allows the user to change the
axis tag without the locations becoming lost.) If not provided, one will be
automatically generated on import from a UUID."""
        },
    )
    min: int = field(
        default=None,
        metadata={
            "description": "The minimum value of this axis, in user space coordinates."
        },
    )
    max: int = field(
        default=None,
        metadata={
            "description": "The maximum value of this axis, in user space coordinates."
        },
    )
    default: int = field(
        default=None,
        metadata={
            "description": """The default value of this axis (center of interpolation),
in user space coordinates. Note that if the min/max/default values are not supplied,
they are returned as `None` in the Python object, and should be computed from the
master locations on export."""
        },
    )
    map: [(int, int)] = field(
        default=None,
        metadata={
            "description": """The mapping between userspace and designspace coordinates."""
        },
    )
    hidden: bool = field(
        default=False,
        metadata={
            "description": """If `True`, this axis is hidden from the user interface.
            Hidden axes are used for internal font generation and are not displayed in the
            user interface."""
        },
    )


@dataclass
class Axis(BaseObject, _AxisFields):
    """Represents an axis in a multiple master or variable font."""

    _write_one_line = True

    def __post_init__(self):
        # If they smacked my name with a bare string, replace with I18NDict
        if isinstance(self.name, str):
            self.name = I18NDictionary.with_default(self.name)
        super().__post_init__()

    def normalize_value(self, value):
        return normalizeValue(
            self.map_forward(value),
            (
                self.map_forward(self.min),
                self.map_forward(self.default),
                self.map_forward(self.max),
            ),
        )

    def denormalize_value(self, value):
        if value == 0:
            return self.default
        elif value > 0:
            return self.default + (self.max - self.default) * value
        else:
            return self.default + (self.default - self.min) * value

    # Compatibility with designspaceLib. Our names are smaller for serialization.
    @property
    def maximum(self):
        return self.max

    @property
    def minimum(self):
        return self.min

    # Compatibility with fontTools.varLib
    @property
    def maxValue(self):
        return self.max

    @property
    def minValue(self):
        return self.min

    @property
    def axisTag(self):
        return self.tag

    @property
    def defaultValue(self):
        return self.default

    @property
    def inverted_map(self) -> Optional[Tuple[float, float]]:
        if not self.map:
            return None
        return [(v, k) for k, v in self.map]

    # Stolen from fontTools.designspaceLib

    def map_forward(self, v):
        from fontTools.varLib.models import piecewiseLinearMap

        if not self.map:
            return v
        return piecewiseLinearMap(v, dict(self.map))

    def map_backward(self, v):
        from fontTools.varLib.models import piecewiseLinearMap

        if not self.map:
            return v
        return piecewiseLinearMap(v, {v: k for k, v in self.map})

    # These are just better names
    def userspace_to_designspace(self, v):
        return self.map_forward(v)

    def designspace_to_userspace(self, v):
        return self.map_backward(v)



from babelfont.Font import Font
from glyphsLib import glyphdata

def zero_mark_widths(font: Font):
    for glyph in font.glyphs:
        glyphinfo = glyphdata.get_glyph(glyph.name)
        if glyphinfo and not (glyphinfo.category == "Mark" and glyphinfo.subCategory == "Nonspacing"):
            continue
        for layer in glyph.layers:
            layer.width = 0



from babelfont.BaseObject import I18NDictionary
import re

from babelfont.Master import CORE_METRICS


def _stash(item, glyphs, include=None, exclude=None):
    if include:
        keys = include
    else:
        keys = glyphs.keys()
    if exclude:
        keys = [k for k in keys if k not in exclude]

    for key in keys:
        if glyphs.get(key):
            if "com.glyphsapp" not in item._formatspecific:
                item._formatspecific["com.glyphsapp"] = {}
            item._formatspecific["com.glyphsapp"][key] = glyphs.get(key)


def _g(item, key, default=None, pop=False):
    if "com.glyphsapp" in item._formatspecific:
        if pop:
            return item._formatspecific["com.glyphsapp"].pop(key, default)
        return item._formatspecific["com.glyphsapp"].get(key, default)
    return default


def _stashed_cp(thing, name):
    return _custom_parameter(thing._formatspecific.get("com.glyphsapp", []), name)


def _custom_parameter(thing, name):
    for param in thing.get("customParameters", []):
        if param["name"] == name:
            return param["value"]
    return None


def _moveformatspecific(item):
    rv = {}
    if "com.glyphsapp" in item._formatspecific:
        rv = {**item._formatspecific.get("com.glyphsapp", {})}
    return rv


def _copyattrs(src, dst, attrs, convertor=lambda x: x):
    for a in attrs:
        v = getattr(src, a)
        if isinstance(v, I18NDictionary):
            v = v.get_default()
        if v:
            dst[a] = convertor(v)


# These are things which are core metrics both for us and glyphs; other
# metrics which we believe to be metrics are stored by Glyphs as custom
# parameters
shared_core_metrics = ["xHeight", "capHeight", "ascender", "descender", "italicAngle"]
custom_parameter_metrics = [x for x in CORE_METRICS if x not in shared_core_metrics]

_rename_metrics = {
    "x-height": "xHeight",
    "cap height": "capHeight",
    "italic angle": "italicAngle",
}
_reverse_rename_metrics = {v: k for k, v in _rename_metrics.items()}


def _glyphs_metrics_to_ours(k):
    return _rename_metrics.get(k, k)


def _our_metrics_to_glyphs(k):
    return _reverse_rename_metrics.get(k, k)


def _metrics_name_to_dict(k):
    m = re.match(r"(.*) \[filter (.*)\]$", k)
    if m:
        return {"type": _our_metrics_to_glyphs(m[1]), "filter": m[2]}
    if k in _reverse_rename_metrics:
        return {"type": _reverse_rename_metrics[k]}
    return {"type": k}


def _metrics_dict_to_name(k):
    name = _glyphs_metrics_to_ours(k.get("type"))
    if "filter" in k:
        name += " [filter " + k.get("filter") + "]"
    return name


def glyphs_i18ndict(i):
    return [{"language": k, "value": v} for k, v in i.items()]


def to_bitfield(flags):
    # Convert an integer flag to an array of bit indices
    return [i for i in range(8) if flags & (1 << i)]


# Taken from glyphsLib
LANGUAGE_MAPPING = {
    "dflt": None,
    "AFK": 0x0436,
    "ARA": 0x0C01,
    "ASM": 0x044D,
    "AZE": 0x042C,
    "BEL": 0x0423,
    "BEN": 0x0845,
    "BGR": 0x0402,
    "BRE": 0x047E,
    "CAT": 0x0403,
    "CSY": 0x0405,
    "DAN": 0x0406,
    "DEU": 0x0407,
    "ELL": 0x0408,
    "ENG": 0x0409,
    "ESP": 0x0C0A,
    "ETI": 0x0425,
    "EUQ": 0x042D,
    "FIN": 0x040B,
    "FLE": 0x0813,
    "FOS": 0x0438,
    "FRA": 0x040C,
    "FRI": 0x0462,
    "GRN": 0x046F,
    "GUJ": 0x0447,
    "HAU": 0x0468,
    "HIN": 0x0439,
    "HRV": 0x041A,
    "HUN": 0x040E,
    "HVE": 0x042B,
    "IRI": 0x083C,
    "ISL": 0x040F,
    "ITA": 0x0410,
    "ITA": 0x0410,
    "IWR": 0x040D,
    "JPN": 0x0411,
    "KAN": 0x044B,
    "KAT": 0x0437,
    "KAZ": 0x043F,
    "KHM": 0x0453,
    "KOK": 0x0457,
    "LAO": 0x0454,
    "LSB": 0x082E,
    "LTH": 0x0427,
    "LVI": 0x0426,
    "MAR": 0x044E,
    "MKD": 0x042F,
    "MLR": 0x044C,
    "MLY": 0x043E,
    "MNG": 0x0352,
    "MTS": 0x043A,
    "NEP": 0x0461,
    "NLD": 0x0413,
    "NOB": 0x0414,
    "ORI": 0x0448,
    "PAN": 0x0446,
    "PAS": 0x0463,
    "PLK": 0x0415,
    "PTG": 0x0816,
    "PTG-BR": 0x0416,
    "RMS": 0x0417,
    "ROM": 0x0418,
    "RUS": 0x0419,
    "SAN": 0x044F,
    "SKY": 0x041B,
    "SLV": 0x0424,
    "SQI": 0x041C,
    "SRB": 0x081A,
    "SVE": 0x041D,
    "TAM": 0x0449,
    "TAT": 0x0444,
    "TEL": 0x044A,
    "THA": 0x041E,
    "TIB": 0x0451,
    "TRK": 0x041F,
    "UKR": 0x0422,
    "URD": 0x0420,
    "USB": 0x042E,
    "UYG": 0x0480,
    "UZB": 0x0443,
    "VIT": 0x042A,
    "WEL": 0x0452,
    "ZHH": 0x0C04,
    "ZHS": 0x0804,
    "ZHT": 0x0404,
}

REVERSE_LANGUAGE_MAPPING = {v: k for v, k in LANGUAGE_MAPPING.items()}


def _to_name_langID(language):
    if language not in LANGUAGE_MAPPING:
        raise ValueError(f"Unknown name language: {language}")
    return LANGUAGE_MAPPING[language]


def _to_glyphs_language(langID):
    if langID not in REVERSE_LANGUAGE_MAPPING:
        raise ValueError(f"Unknown name langID: {langID}")
    return REVERSE_LANGUAGE_MAPPING[langID]


def labels_to_feature(labels):
    feature_names = []
    for label in labels:
        langID = _to_name_langID(label["language"])
        name = label["value"]
        if name == "":
            continue
        name = name.replace("\\", r"\005c").replace('"', r"\0022")
        if langID is None:
            feature_names.append(f'  name "{name}";')
        else:
            feature_names.append(f'  name 3 1 0x{langID:X} "{name}";')
    if feature_names:
        feature_names.insert(0, "featureNames {")
        feature_names.append("};")
    return feature_names



from babelfont import load


# Just testing that it works at all.


def test_ufo_to_glyphs():
    font = load("tests/data/Test1.ufo")
    font.save("tests/data/Test1-from-ufo.glyphs")


# def test_glyphs_to_ufo():
#     font = Babelfont.open("tests/data/Test1.glyphs")
#     Babelfont.save(font, "tests/data/Test1-from-glyphs.ufo")



from io import StringIO
import re
from dataclasses import dataclass, field
from typing import Dict, List, Tuple

from fontTools.feaLib import ast
from fontTools.feaLib.parser import Parser

from .BaseObject import BaseObject

PREFIX_MARKER = "# Prefix: "
PREFIX_RE = re.compile(r"# Prefix: (.*)")


@dataclass
class Features(BaseObject):
    """A representation of the OpenType feature code."""

    classes: Dict[str, List[str]] = field(
        default_factory=dict,
        metadata={
            "separate_items": True,
            "description": "A dictionary of classes. Each group is a list of glyph names or class names. The key should not start with @.",
        },
    )
    prefixes: Dict[str, str] = field(
        default_factory=dict,
        metadata={
            "separate_items": True,
            "description": "A dictionary of OpenType lookups and other feature code to be placed before features are defined. The keys are user-defined names, the values are AFDKO feature code.",
        },
    )
    features: List[Tuple[str, str]] = field(
        default_factory=list,
        metadata={
            "separate_items": True,
            "description": "A list of OpenType feature code, expressed as a tuple (feature tag, code).",
        },
    )

    @classmethod
    def from_fea(cls, fea: str) -> "Features":
        """Load features from a .fea file."""
        parsed = Parser(StringIO(fea), followIncludes=False).parse()
        features = Features()
        currentPrefix = "anonymous"
        for statement in parsed.statements:
            if isinstance(statement, ast.GlyphClassDefinition):
                statement.glyphs.asFea()  # This builds .original
                if statement.glyphs.original:
                    features.classes[statement.name] = [
                        ast.asFea(x) for x in statement.glyphs.original
                    ]
                else:
                    features.classes[statement.name] = list(statement.glyphs.glyphs)
            elif isinstance(statement, ast.FeatureBlock):
                features.features.append((statement.name, ast.Block.asFea(statement)))
            elif isinstance(statement, ast.Comment) and (
                m := re.match(PREFIX_RE, statement.text)
            ):
                currentPrefix = m.group(1)
            else:
                if currentPrefix not in features.prefixes:
                    features.prefixes[currentPrefix] = ""
                features.prefixes[currentPrefix] += statement.asFea()
        return features

    def to_fea(self) -> str:
        """Dump features to a .fea file."""
        fea = ""
        for name, glyphs in self.classes.items():
            fea += f"@{name} = [{' '.join(glyphs)}];\n"
        for prefix, code in self.prefixes.items():
            if prefix != "anonymous":
                fea += f"# Prefix: {prefix}\n"
            fea += code + "\n"
        for name, code in self.features:
            fea += f"feature {name} {{\n{code}}} {name};\n"
        return fea




from babelfont import load, Layer
from fontTools.ttLib import TTFont
import pytest
import glob
import os


def test_load():
    font = load("tests/data/NewFont-Regular.ttf")
    a_layer = font.default_master.get_glyph_layer("A")
    assert isinstance(a_layer, Layer)
    assert len(a_layer.shapes) == 2
    assert len(a_layer.shapes[0].nodes) == 8
    assert len(a_layer.shapes[1].nodes) == 4
    assert a_layer.shapes[0].nodes[0].x == 345
    assert a_layer.shapes[0].nodes[0].y == 700
    assert a_layer.lsb == 25
    assert a_layer.rsb == 25
    assert a_layer.bounds == (25, 0, 505, 700)
    assert a_layer.width == 530

    dot = font.glyphs["uni0307"]
    assert dot.category == "mark"

    assert font.unicode_map[ord("A")] == "A"

    i_layer = font.default_master.get_glyph_layer("i")
    assert isinstance(i_layer, Layer)
    assert len(i_layer.shapes) == 2
    assert i_layer.shapes[0].is_component
    assert i_layer.shapes[0].pos == (0,0)


def test_load_vf():
    font = load("tests/data/Ysabeau[wght].ttf")
    a_layer = font.default_master.get_glyph_layer("A")



from babelfont.Font import Font
from babelfont.Layer import Layer
from babelfont.Anchor import Anchor

def propagate_anchors(font: Font):
    processed = set()

    for glyph in font.glyphs:
        for layer in glyph.layers:
            _propagate_anchors(layer, glyph.name, processed)

def _propagate_anchors(layer: Layer, glyphname: str, processed: set):
    if layer.id in processed:
        return
    processed.add(layer.id)
    
    base_components = []
    mark_components = []
    anchor_names = set()
    to_add = {}
    for component in layer.components:
        component_layer = component.component_layer
        if not component_layer:
            continue
        _propagate_anchors(component_layer, component.ref, processed)
        if any(a.name.startswith("_") for a in component_layer.anchors):
            mark_components.append(component)
        else:
            base_components.append(component)
            anchor_names |= {a.name for a in component_layer.anchors}
    
    if mark_components and not base_components and _is_ligature_mark(layer):
        try:
            component = _component_closest_to_origin(mark_components)
        except Exception as e:
            raise ValueError(
                "Error while determining which component of composite "
                "'{}' is the lowest: {}".format(glyphname, str(e))
            ) from e
        mark_components.remove(component)
        base_components.append(component)
        component_layer = component.component_layer
        anchor_names |= {a.name for a in component_layer.anchors}

    for anchor_name in anchor_names:
        # don't add if parent already contains this anchor OR any associated
        # ligature anchors (e.g. "top_1, top_2" for "top")
        if not any(a.name.startswith(anchor_name) for a in layer.anchors):
            _get_anchor_data(to_add, base_components, anchor_name)

    for component in mark_components:
        _adjust_anchors(to_add, layer, component)

    # we sort propagated anchors to append in a deterministic order
    for name, (x, y) in sorted(to_add.items()):
        layer.anchors.append( Anchor(name=name, x=x, y=y))


def _is_ligature_mark(glyph):
    return not glyph.name.startswith("_") and "_" in glyph.name



from dataclasses import dataclass
from .BaseObject import BaseObject


@dataclass
class _AnchorFields:
    name: str
    x: int = 0
    y: int = 0


@dataclass
class Anchor(BaseObject, _AnchorFields):
    pass



from dataclasses import dataclass, field
from typing import Optional, Union

from .Layer import Layer
from .BaseObject import BaseObject, I18NDictionary
from .Guide import Guide

# Anything which can be varied in MVAR is a master-specific metric
CORE_METRICS = [
    "xHeight",
    "capHeight",
    "ascender",
    "descender",
    "italicAngle",
    "hheaAscender",
    "hheaDescender",
    "hheaLineGap",
    "winAscent",
    "winDescent",
    "typoAscender",
    "typoDescender",
    "typoLineGap",
    "subscriptXSize",
    "subscriptYSize",
    "subscriptXOffset",
    "subscriptYOffset",
    "superscriptXSize",
    "superscriptYSize",
    "superscriptXOffset",
    "superscriptYOffset",
    "strikeoutSize",
    "strikeoutPosition",
    "underlinePosition",
    "underlineThickness",
    "hheaCaretSlopeRise",
    "hheaCaretSlopeRun",
    "hheaCaretOffset",
]


@dataclass
class _MasterFields:
    name: I18NDictionary
    id: str = field(
        repr=False,
        metadata={
            "description": """An ID used to refer to this master in the
`Layer._master` field. (This is allows the user to change the master name
without the layers becoming lost.)"""
        },
    )
    location: dict = field(
        default=None,
        metadata={
            "description": """A dictionary mapping axis tags to coordinates
in order to locate this master in the design space. The coordinates are in designspace units."""
        },
    )
    guides: [Guide] = field(
        default_factory=list,
        repr=False,
        metadata={"separate_items": True, "description": "A list of guides."},
    )
    metrics: dict = field(
        default_factory=dict,
        repr=False,
        metadata={
            "description": """A dictionary mapping metric names (string) to metric value (integer). The following
metric names are reserved: `%s`. Other metrics may be added to this dictionary
as needed by font clients, but their interpretation is not guaranteed to be
compatible between clients."""
            % (",".join(CORE_METRICS))
        },
    )
    kerning: dict = field(  # I think I want this to be UFO-style (l,r) -> value
        default_factory=dict,
        repr=False,
        metadata={
            "separate_items": True,
            "description": "I'll be honest, I haven't worked out how this is meant to work.",
        },
    )
    font: "Font" = field(
        default=None,
        repr=False,
        metadata={
            "python_only": True,
            "description": "Within the Python object, provides a reference to the font object containing this master.",
        },
    )


@dataclass
class Master(BaseObject, _MasterFields):
    """A font master."""

    CORE_METRICS = CORE_METRICS

    def __post_init__(self):
        super().__post_init__()
        # If they smacked my name with a bare string, replace with I18NDict
        if isinstance(self.name, str):
            self.name = I18NDictionary.with_default(self.name)

    def get_glyph_layer(self, glyphname: str) -> Optional[Layer]:
        g = self.font.glyphs[glyphname]
        for layer in g.layers:
            if layer._master == self.id:
                return layer

    @property
    def normalized_location(self) -> dict[str, float]:
        return {a.tag: a.normalize_value(self.location[a.tag]) for a in self.font.axes}

    @property
    def xHeight(self) -> Union[int, float]:
        return self.metrics.get("xHeight", 0)

    @property
    def capHeight(self) -> Union[int, float]:
        return self.metrics.get("capHeight", 0)

    @property
    def ascender(self) -> Union[int, float]:
        return self.metrics.get("ascender", 0)

    @property
    def descender(self) -> Union[int, float]:
        return self.metrics.get("descender", 0)

    @property
    def valid(self) -> bool:
        if not self.font:
            return False
        if self.location and list(self.location.keys()) != [
            n.tag for n in self.font.axes
        ]:
            return False
        return True



from dataclasses import dataclass
from .BaseObject import BaseObject, Color, Position


@dataclass
class _GuideFields:
    pos: Position
    name: str = None
    color: Color = None


@dataclass
class Guide(BaseObject, _GuideFields):
    pass



from babelfont import Axis, Glyph
from io import BytesIO

def test_propagate_format_specific():
    a = Axis(name="Weight", tag="wght", _="Hello")
    assert a._ == "Hello"
    assert a._formatspecific == "Hello"

def test_write_a_negative():
    g = Glyph(name="_test", exported=False)
    s = BytesIO()
    g.write(s)
    assert "exported" in s.getvalue().decode()




from .background import zero_background_width, decompose_backgrounds
from .marks import zero_mark_widths

FILTERS = {
    "zeroBackgroundWidths": zero_background_width,
    "decomposeBackgrounds": decompose_backgrounds,
    "zeroMarkWidths": zero_mark_widths,
}


from babelfont import load


def test_load_many_axes():
    font = load("tests/data/Glyphs2ManyAxes.glyphs")
    assert font.masters[0].location == {'wght': 100, 'wdth': 100, 'opsz': 12, 'ital': 0}
    assert font.masters[1].location == {'wght': 100, 'wdth': 100, 'opsz': 12, 'ital': 1}



from babelfont.convertors.glyphs.glyphs3 import GlyphsThree
import os
import openstep_plist
import glob


class GlyphsPackage(GlyphsThree):
    suffix = ".glyphspackage"

    @classmethod
    def can_save(cls, convertor, **kwargs):
        return False

    @classmethod
    def can_load(cls, other, **kwargs):
        return other.filename.endswith(cls.suffix)

    def _load(self):
        infofile = os.path.join(self.filename, "fontinfo.plist")
        # orderfile = os.path.join(self.filename, "order.plist")
        # glyphorder = openstep_plist.load(open(orderfile, "r"))
        self.scratch["plist"] = openstep_plist.load(
            open(infofile, "r"), use_numbers=True
        )
        self.scratch["plist"]["glyphs"] = []
        for glyphfile in glob.glob(os.path.join(self.filename, "glyphs", "*")):
            self.scratch["plist"]["glyphs"].append(
                openstep_plist.load(open(glyphfile, "r"), use_numbers=True)
            )

        return super()._load()



from babelfont import Font
from babelfont.convertors.designspace import Designspace
from fontTools.designspaceLib import DesignSpaceDocument, SourceDescriptor
import ufoLib2


class UFO(Designspace):
    suffix = ".ufo"

    @classmethod
    def load(cls, convertor):
        self = cls()
        self.ufo = ufoLib2.Font.open(convertor.filename)
        # Wrap it in a DS
        self.ds = DesignSpaceDocument()
        s1 = SourceDescriptor()
        s1.path = convertor.filename
        s1.font = self.ufo
        s1.name = "master.ufo1"
        s1.familyName = self.ufo.info.familyName
        s1.styleName = self.ufo.info.styleName
        self.ds.addSource(s1)
        self.font = Font()
        return self._load()

    def _save(self):
        if len(self.font.masters) > 1:
            raise ValueError("Only single master fonts can be saved as UFO")
        self.save_master_to_ufo(self.font.masters[0], self.filename, is_default=True)



from dataclasses import dataclass, field
from .BaseObject import BaseObject, I18NDictionary


@dataclass
class _InstanceFields:
    name: I18NDictionary = field(metadata={"description": "The name of this instance."})
    location: dict = field(
        metadata={
            "description": """A dictionary mapping axis tags to coordinates in order to locate this instance in the design space."""
        }
    )
    styleName: I18NDictionary = field(
        default_factory=I18NDictionary,
        metadata={"description": "The style name of this instance."},
    )


@dataclass
class Instance(BaseObject, _InstanceFields):
    """An object representing a named or static instance."""

    _write_one_line = True

    def __post_init__(self):
        # If they smacked my name with a bare string, replace with I18NDict
        if isinstance(self.name, str):
            self.name = I18NDictionary.with_default(self.name)
        if isinstance(self.styleName, str):
            self.styleName = I18NDictionary.with_default(self.styleName)
        super().__post_init__()

    @property
    def localisedStyleName(self):
        return self.styleName.as_fonttools_dict or self.name.as_fonttools_dict

    @property
    def postScriptFontName(self):
        return None



import datetime
import math
import re
import uuid
from collections import OrderedDict
from typing import List

import openstep_plist

from babelfont import (
    Anchor,
    Axis,
    Features,
    Glyph,
    Guide,
    Instance,
    Layer,
    Master,
    Node,
    Shape,
    Transform,
)
from babelfont.convertors import BaseConvertor
from babelfont.convertors.glyphs.utils import (
    _copyattrs,
    _custom_parameter,
    _g,
    _glyphs_metrics_to_ours,
    _metrics_dict_to_name,
    _metrics_name_to_dict,
    _moveformatspecific,
    _stash,
    _stashed_cp,
    glyphs_i18ndict,
    labels_to_feature,
    to_bitfield,
    custom_parameter_metrics,
)
from babelfont.Master import CORE_METRICS


class GlyphsThree(BaseConvertor):
    suffix = ".glyphs"

    @classmethod
    def is_suitable_plist(cls, convertor):
        return (
            ".formatVersion" in convertor.scratch["plist"]
            and convertor.scratch["plist"][".formatVersion"] >= 3
        )

    @classmethod
    def can_load(cls, convertor):
        if not super().can_load(convertor):
            return False
        if "plist" not in convertor.scratch:
            convertor.scratch["plist"] = openstep_plist.load(
                open(convertor.filename, "r"), use_numbers=True
            )
        return cls.is_suitable_plist(convertor)

    @classmethod
    def can_save(cls, convertor, **kwargs):
        if not convertor.filename.endswith(".glyphs"):
            return False
        if "format" in kwargs:
            return kwargs["format"] == 3
        return True

    def _load(self):
        font = self.font
        glyphs = self.scratch["plist"]

        for axis in glyphs.pop("axes", []):
            font.axes.append(self.load_axis(axis))

        for master in glyphs.pop("fontMaster"):
            font.masters.append(self.load_master(master))

        for glyph in glyphs.pop("glyphs"):
            font.glyphs.append(self.load_glyph(glyph))

        for instance in glyphs.pop("instances", []):
            instance = self.load_instance(instance)
            if instance:
                font.instances.append(instance)

        self.load_metadata()

        _stash(font, glyphs)

        # Things that start with "interpret" should pop their
        # format-specific values, and will generally need
        # another method to de-interpret them on save

        self.interpret_metrics()
        self.interpret_axes()
        self.interpret_axis_mappings()
        # Oops, we put the axes in designspace coordinates
        for axis in self.font.axes:
            axis.default = axis.designspace_to_userspace(axis.default)
            axis.min = axis.designspace_to_userspace(axis.min)
            axis.max = axis.designspace_to_userspace(axis.max)

        self.interpret_linked_kerning()
        assert self.font.default_master
        self.interpret_kern_groups()
        self.interpret_metric_custom_parameters()
        self.interpret_features()
        return font

    def load_metadata(self):
        glyphs = self.scratch["plist"]
        self.font.upm = glyphs.pop("unitsPerEm")
        self.font.version = (glyphs.pop("versionMajor"), glyphs.pop("versionMinor"))
        self.font.names.familyName.set_default(glyphs.pop("familyName"))

        # This is very glyphs 3
        props = {}
        for prop in glyphs.get("properties", []):
            if "value" in prop:
                props[prop["key"]] = prop["value"]
            else:
                props[prop["key"]] = {p["language"]: p["value"] for p in prop["values"]}

        if props:
            interestingProps = {
                "copyrights": "copyright",
                "designer": "designer",
                "designerURL": "designerURL",
            }  # Etc
            for glyphsname, attrname in interestingProps.items():
                thing = props.get(glyphsname, "")
                if isinstance(thing, dict):
                    getattr(self.font.names, attrname).copy_in(thing)
                else:
                    getattr(self.font.names, attrname).set_default(thing)
            # Do other properties here

        self.font.note = glyphs.pop("note", "")
        self.font.date = datetime.datetime.strptime(
            glyphs.pop("date"), "%Y-%m-%d %H:%M:%S +0000"
        )

    def load_axis(self, gaxis):
        axis = Axis(name=gaxis["name"], tag=gaxis["tag"])
        _stash(axis, gaxis, include=["hidden"])
        return axis

    def load_master(self, gmaster):
        glyphs = self.scratch["plist"]
        master = Master(
            name=gmaster["name"],
            id=gmaster["id"],
            font=self.font,
            guides=[self.load_guide(x) for x in gmaster.pop("guides", [])],
        )
        location = gmaster.pop("axesValues", [])
        master.location = {k.tag: v for k, v in zip(self.font.axes, location)}

        kernmaster = master.id
        kerntable = glyphs.get("kerningLTR", {}).get(kernmaster, {})

        master.kerning = self.load_kerning(kerntable)
        # XXX support RTL kerning etc.

        _stash(master, gmaster)
        assert master.valid
        return master

    def load_glyph(self, gglyph: dict):
        name = gglyph["glyphname"]
        c = gglyph.get("category")
        sc = gglyph.get("subCategory")
        if sc == "Ligature":
            category = "ligature"
        if c == "Mark":
            category = "mark"
        else:
            category = "base"
        cp = self.get_codepoint(gglyph)
        exported = True
        if "export" in gglyph and gglyph.pop("export") == 0:
            exported = False
        g = Glyph(name=name, codepoints=cp or [], category=category, exported=exported)
        g.production_name = gglyph.pop("production", None)

        for layer in gglyph.pop("layers", []):
            g.layers.extend(self.load_layer(layer, g))

        _stash(g, gglyph)

        return g

    def load_layer(self, gslayer: dict, glyph: Glyph, width=None) -> List[Layer]:
        if width is None and "width" in gslayer:
            width = gslayer.pop("width")
        layer = Layer(
            width=width, id=gslayer.pop("layerId", ""), _font=self.font, _glyph=glyph
        )
        layer.vertWidth = gslayer.pop("vertWidth", None)
        layer.name = gslayer.pop("name", "")
        if [x for x in self.font.masters if x.id == layer.id]:
            layer._master = layer.id
        else:
            layer._master = gslayer.pop("associatedMasterId", None)
        layer.guides = [
            self.load_guide(x)
            for x in gslayer.pop("guideLines", gslayer.pop("guides", []))
        ]
        layer.shapes = []
        for shape in gslayer.pop("shapes", []):
            layer.shapes.append(self.load_shape(shape))
        for shape in gslayer.pop("paths", []):
            layer.shapes.append(self.load_path(shape))
        for shape in gslayer.pop("components", []):
            layer.shapes.append(self.load_component(shape))
        for anchor in gslayer.pop("anchors", []):
            layer.anchors.append(self.load_anchor(anchor))

        returns = [layer]
        layer.glyph = glyph
        if "background" in gslayer:
            (background,) = self.load_layer(
                gslayer["background"], glyph, width=layer.width
            )
            # If it doesn't have an ID, we need to generate one
            background.id = background.id or str(uuid.uuid1())
            background.isBackground = True

            layer.background = background.id
            returns.append(background)
        for r in returns:
            assert r.valid

        _stash(layer, gslayer)
        return returns

    def load_guide(self, gguide):
        g = Guide(
            pos=[*gguide.get("pos", (0, 0)), gguide.get("angle", 0)],
            name=gguide.get("name", None),
        )
        if gguide.get("locked"):
            if g.name is None:
                g.name = ""
            g.name += " [locked]"
        _stash(g, gguide)
        return g

    def load_anchor(self, ganchor):
        x, y = ganchor.get("pos", (0, 0))
        return Anchor(name=ganchor["name"], x=x, y=y)

    def load_component(self, shape):
        glyphname = shape.pop("ref")
        transform = shape.pop("transform", None)
        if isinstance(transform, str):
            m = re.match(r"^\{(\S+), (\S+), (\S+), (\S+), (\S+), (\S+)\}", transform)
            transform = Transform(*[float(g) for g in m.groups()])
        c = Shape(ref=glyphname)

        if not transform:
            translate = Transform().translate(*shape.pop("pos", (0, 0)))
            scale = Transform().scale(*shape.pop("scale", (1, 1)))
            rotation = Transform().rotate(math.radians(shape.pop("angle", 0)))
            # Compute transform...
            transform = translate.transform(scale).transform(rotation)

        c.transform = transform
        _stash(c, shape)
        return c

    def load_shape(self, shape):
        if "nodes" in shape:  # It's a path
            return self.load_path(shape)
        else:
            return self.load_component(shape)

    def load_instance(self, ginstance):
        if ginstance.get("type") == "variable":
            return
        if "axesValues" in ginstance:
            location = ginstance.pop("axesValues")
            instance_location = {k.tag: v for k, v in zip(self.font.axes, location)}
        elif "instanceInterpolations" in ginstance:
            # All right then.
            instance_location = {k.tag: 0 for k in self.font.axes}
            for mId, factor in ginstance.get("instanceInterpolations").items():
                master_loc = self.font.master(mId).location
                for k in self.font.axes:
                    instance_location[k.tag] += master_loc[k.tag] * factor
        else:
            raise ValueError("Need to Synthesize location")
        name = ginstance.pop("name")
        i = Instance(
            name=name,
            styleName=name,
            location=instance_location,
        )
        _stash(i, ginstance)
        return i

    def load_path(self, path):
        shape = Shape()
        shape.nodes = [Node(*n) for n in path["nodes"]]
        shape.closed = path["closed"]
        _stash(shape, path)
        if len(shape.nodes):
            # Bring it to front
            shape.nodes = shape.nodes[-1:] + shape.nodes[:-1]
        return shape

    def get_codepoint(self, gglyph):
        cp = gglyph.get("unicode")
        if cp and not isinstance(cp, list):
            cp = [cp]
        if cp:
            return [int(x) for x in cp]

    def interpret_kern_groups(self):
        for g in self.font.glyphs:
            left_group = _g(g, "kernLeft", pop=True)
            if left_group:
                self.font.second_kern_groups.setdefault(left_group, []).append(g.name)

            right_group = _g(g, "kernRight", pop=True)
            if right_group:
                self.font.first_kern_groups.setdefault(right_group, []).append(g.name)

    def interpret_linked_kerning(self):
        name_to_master = {m.name.get_default(): m for m in self.font.masters}
        for master in self.font.masters:
            kernmaster = None
            if _stashed_cp(master, "Link Metrics With First Master"):
                master.kerning = self.font.masters[0].kerning
            elif _stashed_cp(master, "Link Metrics With Master"):
                kernmaster_name = _stashed_cp(master, "Link Metrics With Master")
                kernmaster = name_to_master.get(kernmaster_name)
                if kernmaster:
                    master.kerning = kernmaster.kerning

    def load_kerning(self, kerndict):
        def ungroup(groupname):
            if groupname[0] != "@":
                return groupname
            return groupname.replace("@MMK_L", "@first_group").replace(
                "@MMK_R", "@second_group"
            )

        return {
            (ungroup(left), ungroup(right)): value
            for left, level2 in kerndict.items()
            for right, value in level2.items()
        }

    def interpret_metrics(self):
        metrics = _g(self.font, "metrics", [])  # Keep stashed
        metric_types = [_metrics_dict_to_name(k) for k in metrics]
        for master in self.font.masters:
            metric_values = _g(master, "metricValues", [], pop=True)
            master.metrics["italicAngle"] = 0
            for k, v in zip(metric_types, metric_values):
                pos = v.get("pos", 0)
                master.metrics[k] = pos
                if v.get("over"):
                    master.metrics["%s overshoot" % _glyphs_metrics_to_ours(k)] = v[
                        "over"
                    ]

    def interpret_axes(self):
        # The default master in glyphs is either the first master or the
        # one selected by the Variable Font Origin custom parameter
        vfo = _stashed_cp(self.font, "Variable Font Origin")
        if not vfo:
            vfo = self.font.masters[0].id

        for master in self.font.masters:
            for axis in self.font.axes:
                thisLoc = master.location[axis.tag]
                if axis.min is None or thisLoc < axis.min:
                    axis.min = thisLoc
                if master.id == vfo:
                    axis.default = master.location[axis.tag]
                if axis.max is None or thisLoc > axis.max:
                    axis.max = thisLoc

    def interpret_axis_mappings(self):
        # There may be a "Axis Mappings" custom parameter, if so use that
        axes_by_name = {a.name.get_default(): a for a in self.font.axes}

        if mapping := _stashed_cp(self.font, "Axis Mappings"):
            axes_by_tag = {a.tag: a for a in self.font.axes}
            for axistag, axismap in mapping.items():
                ax = axes_by_tag[axistag]
                ax.map = []
                for designspace, userspace in axismap.items():
                    ax.map.append((userspace, float(designspace)))
            return

        for instance in self.font.instances:
            c = (
                _custom_parameter(
                    instance._formatspecific.get("com.glyphsapp", {}), "Axis Location"
                )
                or []
            )
            for loc in c:
                ax = axes_by_name[loc["Axis"]]
                if not ax.map:
                    ax.map = []
                ax.map.append(
                    (
                        instance.location[ax.tag],
                        int(loc["Location"]),
                    )
                )

    def interpret_metric_custom_parameters(self):
        for master in self.font.masters:
            cps = _g(master, "customParameters", [])
            new_cps = []
            for param in cps:
                if param["name"] in CORE_METRICS:
                    master.metrics[param["name"]] = param["value"]
                else:
                    new_cps.append(param)
            _stash(master, {"customParameters": new_cps})

    def interpret_features(self):
        self.font.features = Features()

        def code_with_notes(feature):
            code = feature["code"]
            if "disabled" in feature:
                code = "# Disabled\n# " + code.replace("\n", "\n# ")
            if "automatic" in feature:
                code = "# Automatic\n" + code
            if "labels" in feature:
                code += "\n".join(labels_to_feature(feature["labels"]))

            return code

        for glyphclass in _g(self.font, "classes", [], pop=True):
            # Beware of tokens XXX
            self.font.features.classes[glyphclass["name"]] = glyphclass["code"].split()
        for prefix in _g(self.font, "featurePrefixes", [], pop=True):
            self.font.features.prefixes[prefix["name"]] = code_with_notes(prefix)
        for feature in _g(self.font, "features", [], pop=True):
            self.font.features.features.append(
                (feature["tag"], code_with_notes(feature))
            )

    def _save(self):
        font = self.font
        out = _moveformatspecific(font)
        self.glyphs = out
        out["versionMajor"], out["versionMinor"] = font.version
        out[".formatVersion"] = 3
        out["unitsPerEm"] = font.upm
        if font.note:
            out["note"] = font.note
        if font.date:
            out["date"] = font.date.strftime("%Y-%m-%d %H:%M:%S +0000")
        out["familyName"] = font.names.familyName.default_or_dict()
        out["axes"] = [self.save_axis(ax) for ax in self.font.axes]

        # Sort out the metrics order, using "my" names
        metrics_order = _g(font, "metrics", [])
        # Ensure we have all the metrics
        for m in font.masters:
            for k in m.metrics.keys():
                if k in custom_parameter_metrics:
                    continue
                if k.endswith(" overshoot"):
                    continue  # We'll write it into the other metric
                k = _metrics_name_to_dict(k)
                if k not in metrics_order:
                    metrics_order.append(k)

        out["metrics"] = metrics_order

        out["fontMaster"] = [self.save_master(m) for m in self.font.masters]
        out["glyphs"] = [self.save_glyph(g) for g in self.font.glyphs]
        instances = [self.save_instance(i) for i in self.font.instances]
        if instances:
            out["instances"] = instances

        kerntables = OrderedDict()
        for master in self.font.masters:
            table = self.save_kerning(master.kerning)
            if table:
                kerntables[master.id] = table
        if kerntables:
            out["kerningLTR"] = kerntables

        self.save_metadata(out)
        self.save_custom_parameters(out)
        self.save_features(out)

        with open(self.filename, "wb") as file:
            openstep_plist.dump(out, file, indent=0, single_line_tuples=True)
            file.write(b"\n")

    def save_axis(self, axis):
        gaxis = _moveformatspecific(axis)
        _copyattrs(axis, gaxis, ["name", "tag"])
        return gaxis

    def save_kerning(self, kerntable):
        newtable = {}
        for (left, right), val in kerntable.items():
            left = left.replace("@first_group", "@MMK_L").replace(
                "@second_group", "@MMK_R"
            )
            right = right.replace("@first_group", "@MMK_L").replace(
                "@second_group", "@MMK_R"
            )
            newtable.setdefault(left, {})[right] = val
        return newtable

    def save_master(self, master):
        gmaster = _moveformatspecific(master)
        if master.location:
            gmaster["axesValues"] = list(master.location.values())
        gmaster["metricValues"] = []
        if "customParameters" not in gmaster:
            gmaster["customParameters"] = []
        for metric in custom_parameter_metrics:
            if metric in self.font.default_master.metrics:
                gmaster["customParameters"].append(
                    {
                        "name": metric,
                        "value": self.font.default_master.metrics[metric],
                    }
                )
        if not gmaster["customParameters"]:
            del gmaster["customParameters"]

        for k in self.glyphs["metrics"]:
            metric = {}
            pos = master.metrics.get(_metrics_dict_to_name(k))
            over = master.metrics.get(_metrics_dict_to_name(k) + " overshoot")
            if pos:
                metric["pos"] = pos
            if over:
                metric["over"] = over
            gmaster["metricValues"].append(metric)
        if master.guides:
            gmaster["guides"] = [self.save_guide(g) for g in master.guides]
        _copyattrs(master, gmaster, ["name", "id"], convertor=str)
        return gmaster

    def save_guide(self, guide):
        gguide = _moveformatspecific(guide)
        gguide["pos"] = tuple(guide.pos[0:2])
        if guide.pos[2]:
            gguide["angle"] = guide.pos[2]
        return gguide

    def save_glyph(self, glyph):
        gglyph = _moveformatspecific(glyph)
        gglyph["glyphname"] = glyph.name
        if len(glyph.codepoints) == 1:
            # if glyph.codepoints[0] < 256:
            gglyph["unicode"] = glyph.codepoints[0]
            # else:
            # gglyph["unicode"] = "%04x" % glyph.codepoints[0]
        elif len(glyph.codepoints) > 1:
            gglyph["unicode"] = glyph.codepoints
        gglyph["layers"] = [
            self.save_layer(layer) for layer in glyph.layers if not layer.isBackground
        ]
        if glyph.production_name is not None and glyph.production_name != glyph.name:
            gglyph["production"] = glyph.production_name
        if not glyph.exported:
            gglyph["export"] = 0
        # Check for kern groups
        for group, members in self.font.second_kern_groups.items():
            if glyph.name in members:
                gglyph["kernLeft"] = group
        for group, members in self.font.first_kern_groups.items():
            if glyph.name in members:
                gglyph["kernRight"] = group
        # XXX
        return gglyph

    def save_layer(self, layer):
        glayer = _moveformatspecific(layer)
        _copyattrs(layer, glayer, ["width", "name"])
        if layer.vertWidth is not None:
            glayer["vertWidth"] = layer.vertWidth
        glayer["layerId"] = str(layer.id)
        if layer.guides:
            glayer["guides"] = [self.save_guide(g) for g in layer.guides]
        if layer.shapes:
            glayer["shapes"] = [self.save_shape(s) for s in layer.shapes]
        if layer._master and layer._master != layer.id:
            glayer["associatedMasterId"] = layer._master
        if layer.anchors:
            glayer["anchors"] = [self.save_anchor(a) for a in layer.anchors]
        if layer.background:
            glayer["background"] = self.save_layer(layer._background_layer())
            del glayer["background"]["width"]
            del glayer["background"]["layerId"]
        return glayer

    def save_shape(self, shape):
        gshape = _moveformatspecific(shape)
        if shape.is_path:
            gshape["closed"] = shape.closed
            outputnodes = []
            if len(shape.nodes):
                # Put front back at end
                outputnodes = shape.nodes[1:] + [shape.nodes[0]]

            gshape["nodes"] = [self.save_node(n) for n in outputnodes]
        else:
            _copyattrs(shape, gshape, ["ref", "angle"])
            if shape.pos != (0, 0):
                _copyattrs(shape, gshape, ["pos"])
            if shape.scale != (1, 1):
                _copyattrs(shape, gshape, ["scale"])
        return gshape

    def save_node(self, node):
        if node.userdata:
            return (node.x, node.y, node.type, node.userdata)
        else:
            return (node.x, node.y, node.type)

    def save_instance(self, instance):
        ginstance = _moveformatspecific(instance)
        ginstance["name"] = instance.name.get_default()
        ginstance["axesValues"] = [instance.location[ax.tag] for ax in self.font.axes]

        return ginstance

    def save_anchor(self, anchor):
        ganchor = _moveformatspecific(anchor)
        ganchor["name"] = anchor.name
        ganchor["pos"] = (anchor.x, anchor.y)
        return ganchor

    def save_metadata(self, out):
        if self.font.note:
            out["note"] = self.font.note
        if self.font.version:
            out["versionMajor"], out["versionMinor"] = self.font.version
        if "properties" not in out:
            out["properties"] = []
        props = out["properties"]
        alreadydone = {p["key"] for p in props}
        if self.font.names.copyright and "copyrights" not in alreadydone:
            props.append(
                {
                    "key": "copyrights",
                    "values": glyphs_i18ndict(self.font.names.copyright),
                }
            )
        if self.font.names.designer and "designer" not in alreadydone:
            props.append(
                {"key": "designer", "values": glyphs_i18ndict(self.font.names.designer)}
            )
        if self.font.names.designerURL and "designerURL" not in alreadydone:
            props.append(
                {
                    "key": "designerURL",
                    "values": glyphs_i18ndict(self.font.names.designerURL),
                }
            )
        if not props:
            del out["properties"]

    def save_custom_parameters(self, out):
        if "customParameters" not in out:
            out["customParameters"] = []
        for otvalue in self.font.customOpenTypeValues:
            table, field, value = otvalue.table, otvalue.field, otvalue.value
            if table == "OS/2" and field == "fsType":
                out["customParameters"].append(
                    {"name": "fsType", "value": to_bitfield(int(value))}
                )
            if table == "OS/2" and field == "fsSelection":
                if value & 0x7:
                    out["customParameters"].append(
                        {"name": "Use Typo Metrics", "value": 1}
                    )
                if value & 0x8:
                    out["customParameters"].append(
                        {"name": "Has WWS Names", "value": 1}
                    )
        if not out["customParameters"]:
            del out["customParameters"]

    def save_features(self, out):
        # In truth we should probably steal glyphsLib.builder.features._process_feature_block
        def _save_item(key, value, name):
            item = {}
            item[name] = key
            if value.startswith("# Automatic\n"):
                item["automatic"] = True
                value = re.sub("^# Automatic\n", "", value, flags=re.MULTILINE)
            if value.startswith("# Disabled\n"):
                item["disabled"] = True
                value = re.sub("^#", "", value[10:], flags=re.MULTILINE)
            item["code"] = value
            return item

        if self.font.features:
            out["classes"] = [
                {"name": k, "code": " ".join(v)}
                for k, v in self.font.features.classes.items()
            ]
            out["featurePrefixes"] = [
                _save_item(k, v, "name") for k, v in self.font.features.prefixes.items()
            ]
            out["features"] = [
                _save_item(k, v, "tag") for k, v in self.font.features.features
            ]
        else:
            out["classes"] = []
            out["featurePrefixes"] = []
            out["features"] = []
        if not out["classes"]:
            del out["classes"]
        if not out["featurePrefixes"]:
            del out["featurePrefixes"]
        if not out["features"]:
            del out["features"]



from babelfont import load

def test_load_glyph():
    filename = "tests/data/Castoro Roman.vfj"
    font = load(filename)

    assert len(font.default_master.get_glyph_layer("A").paths[0].nodes) == 50
    assert len(font.default_master.get_glyph_layer("A").paths[1].nodes) == 4




from babelfont import load
import defcon


def test_ufo_load():
    font = load("tests/data/Test1.ufo")
    assert font.unicode_map[ord("A")] == "A"

# Can't save UFO right now
# def test_ufo_roundtrip():
#     font = load("tests/data/Test1.ufo")
#     font.save("tests/data/Test1.roundtrip.ufo")
#     dc1 = defcon.Font("tests/data/Test1.ufo")
#     dc2 = defcon.Font("tests/data/Test1.roundtrip.ufo")
#     assert dc1.getDataForSerialization() == dc2.getDataForSerialization()



from babelfont.convertors import Convert
from babelfont.fontFilters import FILTERS
import argparse
import sys


def main():
    parser = argparse.ArgumentParser(
        prog="babelfont", description="Convert between font formats"
    )
    parser.add_argument(
        "--filter",
        "-f",
        help="Filter to apply",
        action="append",
        choices=FILTERS.keys(),
    )
    parser.add_argument("input", metavar="IN", help="Input file")
    parser.add_argument("output", metavar="OUT", help="Output file")
    args = parser.parse_args()

    try:
        font = Convert(args.input).load()
    except Exception as e:
        print("Couldn't read %s: %s" % (args.input, e))
        raise e
        sys.exit(1)

    for filter in args.filter or []:
        if filter not in FILTERS:
            print("Unknown filter %s" % filter)
            continue
        FILTERS[filter](font)

    try:
        Convert(args.output).save(font)
    except Exception as e:
        print("Couldn't write %s: %s" % (args.output, e))
        raise e
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    sys.exit(main())



from babelfont import Features


def test_parse():
    features = Features.from_fea("""
lookup foobar { 
    sub a by b;                                 
} foobar;
@group_one = [a b c];
@group_two = [@group_one d e f];
feature test {
    sub X by Y;
} test;
lookup foobar2 { sub c by d; } foobar2;
# Prefix: namedPrefix
lookup foobar3 {
    sub e by f;                                 
} foobar3;
feature test {
    pos X Y -150;
} test;
""")
    assert "group_one" in features.classes
    assert features.classes["group_one"] == ["a", "b", "c"]
    assert features.classes["group_two"] == ["@group_one", "d", "e", "f"]
    assert features.prefixes["anonymous"] == "lookup foobar {\n    sub a by b;\n} foobar;\nlookup foobar2 {\n    sub c by d;\n} foobar2;\n"
    assert len(features.features) == 2
    assert features.prefixes["namedPrefix"] == "lookup foobar3 {\n    sub e by f;\n} foobar3;\n"
    assert features.features[0] == ("test", "    sub X by Y;\n")
    assert features.features[1] == ("test", "    pos X Y -150;\n")



import re
import uuid
import logging
from pathlib import Path

from babelfont.BaseObject import OTValue
from babelfont.convertors import BaseConvertor

from babelfont import (
    Master,
    Instance,
    Glyph,
    Layer,
    Shape,
    Transform,
)

NUMBER = r"(-?\d+(?:\.\d+)?(?:e-?\d+)?)"
MOVE_RE = re.compile(rf"^\s*{NUMBER}\s+{NUMBER}\s+m")
CURVE_RE = re.compile(
    rf"^\s*{NUMBER}\s+{NUMBER}\s+{NUMBER}\s+{NUMBER}\s+{NUMBER}\s+{NUMBER} c"
)
LINE_RE = re.compile(rf"^\s*{NUMBER}\s+{NUMBER}\s+l")

log = logging.getLogger(__name__)


class FontForgeSFDIR(BaseConvertor):
    suffix = ".sfdir"
    IGNORE = [
        "SplineFontDB",
        "UComments",
        "TeXData",
        "FontName",
        "DefaultBaseFilename",
        "sfntRevision",
        "InvalidEm",
        "XUID",
        "OS2Version",
        "PfmFamily",
        "AntiAlias",
        "FitToEm",
        "DisplaySize",
        "Validated",
        "Flags",
        "Lookup",
    ]

    @classmethod
    def can_save(cls, convertor, **kwargs):
        return False

    def _load(self):
        font = self.font
        self.current_glyph = None
        # I'm going to assume a single master font
        font.masters = [Master(name="Regular", id=str(uuid.uuid4()), font=font)]
        font.instances = [Instance(name="Regular", styleName="Regular", location={})]

        self.info = open(Path(self.filename) / "font.props").read().splitlines()
        self._load_line("info")
        # Probe once for glyphorder
        self.glyph_order = {}
        for fontfile in Path(self.filename).glob("*.glyph"):
            self._probe_glyphorder(fontfile)
        for gid in sorted(self.glyph_order.keys()):
            self.font.glyphs.append(Glyph(name=self.glyph_order[gid]))

        for fontfile in Path(self.filename).glob("*.glyph"):
            self.info = open(fontfile).read().splitlines()
            self._load_line("glyph")

        return font

    def _probe_glyphorder(self, file):
        glyphname = ""
        for line in open(file).readlines():
            if m := re.match(r"StartChar: (.*)", line):
                glyphname = m.group(1)
                continue
            if m := re.match(r"Encoding: (\d+) (-?\d+) (\d+)", line):
                gid = int(m.group(3))
                if not glyphname:
                    raise ValueError(f"Bad file {file} - no StartChar")
                self.glyph_order[gid] = glyphname
                break

    def _load_line(self, mode):
        while self.info:
            line = self.info.pop(0)
            if line == "EndSplineFont" and mode == "info":
                break
            if line == "EndChar" and mode == "glyph":
                break
            if ": " not in line:
                if hasattr(self, "_handle_" + line):
                    getattr(self, "_handle_" + line)()
                    continue
                else:
                    raise ValueError("Bad font.props file - " + line)
            key, value = line.split(": ", 1)
            if key in self.IGNORE:
                continue
            if hasattr(self, "_handle_" + key):
                getattr(self, "_handle_" + key)(value)
            else:
                log.warn("Unknown info key %s", key)

    def _handle_FamilyName(self, value):
        self.font.names.familyName.set_default(value)

    def _handle_FullName(self, value):
        self.font.names.compatibleFullName.set_default(value)

    def _handle_Copyright(self, value):
        self.font.names.copyright.set_default(value)

    def _handle_Comments(self, value):
        self.font.notes = value

    def _handle_Version(self, value):
        self.font.names.version.set_default(value)
        self.font.version = [int(x) for x in value.split(".")]

    def _handle_Weight(self, value):
        self.font.names.typographicSubfamily.set_default(value)

    def _handle_Fontlog(self, value):
        self.font.names.description.set_default(value)

    def _handle_ItalicAngle(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("post", "italicAngle", float(value))
        )

    def _handle_UnderlinePosition(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("post", "underlinePosition", int(value))
        )

    def _handle_UnderlineWidth(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("post", "underlineThickness", int(value))
        )

    def _handle_OS2TypoAscent(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("OS/2", "sTypoAscender", int(value))
        )

    def _handle_OS2TypoDescent(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("OS/2", "sTypoDescender", int(value))
        )

    def _handle_OS2TypoLineGap(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("OS/2", "sTypoLineGap", int(value))
        )

    def _handle_OS2WinAscent(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("OS/2", "usWinAscent", int(value))
        )

    def _handle_OS2WinDescent(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("OS/2", "usWinDescent", int(value))
        )

    def _handle_OS2Vendor(self, value):
        self.font.customOpenTypeValues.append(OTValue("OS/2", "achVendID", value))

    def _handle_hheaAscent(self, value):
        self.font.customOpenTypeValues.append(OTValue("hhea", "ascent", int(value)))

    def _handle_hheaDescent(self, value):
        self.font.customOpenTypeValues.append(OTValue("hhea", "descent", int(value)))

    def _handle_OS2TypoLinegap(self, value):
        self.font.customOpenTypeValues.append(
            OTValue("OS/2", "sTypoLineGap", int(value))
        )

    def _handle_Ascent(self, value):
        self.font.masters[0].metrics["ascender"] = int(value)

    def _handle_Descent(self, value):
        self.font.masters[0].metrics["descender"] = -int(value)

    def _handle_LayerCount(self, value):
        # Ignore layers for now
        pass

    def _handle_Layer(self, value):
        # Ignore layers for now
        pass

    def _handle_FSType(self, value):
        self.font.customOpenTypeValues.append(OTValue("OS/2", "fsType", int(value)))

    def _handle_OS2_WeightWidthSlopeOnly(self, value):
        # XXX
        self.font.customOpenTypeValues.append(
            OTValue("OS/2", "fsSelection", int(value))
        )

    def _handle_OS2_UseTypoMetrics(self, value):
        if not value:
            return
        for otval in self.font.customOpenTypeValues:
            if otval.table == "OS/2" and otval.field == "fsSelection":
                otval.value = otval.value | 7 << 1
                return
        self.font.customOpenTypeValues.append(OTValue("OS/2", "fsSelection", 7))

    def _handle_CreationTime(self, value):
        self.font.customOpenTypeValues.append(OTValue("head", "created", int(value)))

    def _handle_ModificationTime(self, value):
        self.font.customOpenTypeValues.append(OTValue("head", "modified", int(value)))

    def _handle_ShortTable(self, value):
        # Ignore short table for now
        while self.info:
            line = self.info.pop(0)
            if line.startswith("EndShort"):
                break

    def _handle_TtTable(self, value):
        while self.info:
            line = self.info.pop(0)
            if line.startswith("EndTTInstrs"):
                break

    def _handle_BeginPrivate(self, value):
        while self.info:
            line = self.info.pop(0)
            if line.startswith("EndPrivate"):
                break

    def _handle_Grid(self):
        grid = Layer(name="grid", id="grid", _master="default", _font=self.font)
        self._expect_splineset(grid.getPen())

    def _handle_StartChar(self, value):
        self.current_glyph = self.font.glyphs[value]
        self.current_glyph.layers = [
            Layer(
                name="default",
                id=self.font.masters[0].id,
                _master=self.font.masters[0].id,
                _font=self.font,
                _glyph=self.current_glyph,
            )
        ]

    def _handle_Encoding(self, value):
        if not self.current_glyph:
            return
        _old, unicode, gid = value.split()
        if unicode != "-1":
            self.current_glyph.codepoints.append(int(unicode))

    def _handle_Width(self, value):
        self.current_glyph.layers[0].width = int(value)

    def _handle_VWidth(self, value):
        self.current_glyph.layers[0].height = int(value)

    def _handle_GlyphClass(self, value):
        pass

    def _handle_Fore(self):
        self._expect_splineset(self.current_glyph.layers[0].getPen())

    def _handle_Refer(self, value: str):
        gid, unicode, style, xx, xy, yx, yy, dx, dy, ttflag = value.split(None, 10)
        component = Shape(
            ref=self.glyph_order[int(gid)],
            transform=Transform(
                float(xx), float(xy), float(yx), float(yy), float(dx), float(dy)
            ),
        )
        self.current_glyph.layers[0].shapes.append(component)

    def _expect_splineset(self, pen):
        paths = False
        while self.info:
            line = self.info.pop(0)
            if line == "SplineSet":
                continue
            if line.startswith("EndSplineSet"):
                break
            if m := re.match(MOVE_RE, line):
                if paths:
                    pen.closePath()
                pen.moveTo((float(m.group(1)), float(m.group(2))))
            elif m := re.match(CURVE_RE, line):
                if (float(m.group(1)), float(m.group(2))) == (
                    float(m.group(3)),
                    float(m.group(4)),
                ):
                    # Quadratic in disguise
                    pen.qCurveTo(
                        (float(m.group(1)), float(m.group(2))),
                        (float(m.group(5)), float(m.group(6))),
                    )
                else:
                    pen.curveTo(
                        (float(m.group(1)), float(m.group(2))),
                        (float(m.group(3)), float(m.group(4))),
                        (float(m.group(5)), float(m.group(6))),
                    )
                paths = True
            elif m := re.match(LINE_RE, line):
                pen.lineTo((float(m.group(1)), float(m.group(2))))
                paths = True
            else:
                # Put back
                self.info.insert(0, line)
                break
        if paths:
            pen.closePath()
        return



import openstep_plist
from babelfont import load
import pytest
import os
import re


def normalize(plist):
    return openstep_plist.dumps(
        openstep_plist.loads(plist, use_numbers=True), indent=0, single_line_tuples=True
    )


def data_path(filename):
    path, _ = os.path.split(__file__)
    return os.path.join(path, "data", filename)


def check_equal(one, two):
    with open(one) as f:
        expected = f.read()
    with open(two) as f:
        got = f.read()
    assert normalize(got) == normalize(expected)


@pytest.mark.parametrize(
    "filename", ["SimpleTwoAxis3.glyphs", "GlyphsFileFormatv3.glyphs"]
)
def test_glyphs3_babelfont_glyphs3(tmp_path, filename):
    original_file = data_path(filename)
    tmp_file = os.path.join(tmp_path, "Roundtrip.babelfont")
    roundtrip_file = os.path.join(tmp_path, "Roundtrip.glyphs")
    f = load(original_file)
    f.save(tmp_file)
    # f2 = load(tmp_file)
    # # Remove the metrics, reconstruct them
    # del f2._formatspecific["com.glyphsapp"]["metrics"]
    f.save(roundtrip_file, format=3)
    check_equal(original_file, roundtrip_file)



from dataclasses import dataclass, field
from functools import cached_property
from typing import List, Optional

from fontTools.ufoLib.pointPen import (
    PointToSegmentPen,
    SegmentToPointPen,
    AbstractPointPen,
)
from fontTools.pens.boundsPen import BoundsPen
from fontTools.pens.recordingPen import DecomposingRecordingPen

from .BaseObject import BaseObject, Color
from .Guide import Guide
from .Anchor import Anchor
from .Node import Node
from .Shape import Shape
import uuid


@dataclass
class _LayerFields:
    width: int = 0
    height: int = 0
    vertWidth: Optional[int] = None
    name: str = None
    _master: str = None
    id: str = field(default_factory=lambda: str(uuid.uuid1()))
    guides: List[Guide] = field(default_factory=list, repr=False)
    shapes: List[Shape] = field(
        default_factory=list, repr=False, metadata={"separate_items": True}
    )
    anchors: List[Anchor] = field(default_factory=list, repr=False)
    color: Color = None
    layerIndex: int = 0
    # hints: List[Hint]
    background: Optional[str] = field(default=None, repr=False)
    isBackground: bool = field(default=False, repr=False)
    location: List[float] = None
    _font: Optional["Font"] = field(
        default=None, repr=False, metadata={"python_only": True}
    )
    _glyph: Optional["Glyph"] = field(
        default=None, repr=False, metadata={"python_only": True}
    )


@dataclass
class Layer(BaseObject, _LayerFields):
    @property
    def master(self):
        assert self._font
        if not self._master:
            return None
        return self._font.master(self._master)

    @property
    def paths(self) -> List[Shape]:
        return [x for x in self.shapes if x.is_path]

    @property
    def components(self) -> List[Shape]:
        return [x for x in self.shapes if x.is_component]

    def recursive_component_set(self):
        mine = set([x.ref for x in self.components])
        theirs = set()
        for c in mine:
            other_layer = self.master.get_glyph_layer(c)
            theirs |= other_layer.recursive_component_set()
        return mine | theirs

    def _background_of(self) -> Optional["Layer"]:
        for layer in self._glyph.layers:
            if layer.background == self.id:
                return layer

    def _background_layer(self) -> Optional["Layer"]:
        if not self.background:
            return
        for layer in self._glyph.layers:
            if layer.id == self.background:
                return layer

    def _nested_component_dict(self) -> dict[str, "Layer"]:
        result = {}
        todo = [x.ref for x in self.components]
        while todo:
            current = todo.pop()
            if current in result:
                continue
            if self.master:
                result[current] = self.master.get_glyph_layer(current)
            else:
                # Find a glyph with same layerid?
                for layer in self._font.glyphs[current].layers:
                    if layer.id == self.id:
                        result[current] = layer
                        break
                if current not in result and self.isBackground:
                    master_layer = self._background_of()
                    if master_layer:
                        master = master_layer._font.master(master_layer._master)
                        result[current] = master.get_glyph_layer(current)
                        if result[current] and result[current]._background_layer():
                            result[current] = result[current]._background_layer()

                if current not in result or not result[current]:
                    raise ValueError("Could not find layer")
            todo.extend([x.ref for x in result[current].components])
        return result

    @cached_property
    def bounds(self):
        glyphset = {}
        for c in list(self.recursive_component_set()):
            glyphset[c] = self.master.get_glyph_layer(c)
        pen = BoundsPen(glyphset)
        self.draw(pen)
        return pen.bounds

    @property
    def lsb(self):
        if not self.bounds:  # Space glyph
            return 0
        return self.bounds[0]

    @property
    def rsb(self):
        if not self.bounds:  # Space glyph
            return 0
        return self.width - self.bounds[2]

    @property
    def valid(self):
        if not self._font or not self._glyph:
            return False
        return True

    @property
    def anchors_dict(self):
        return {a.name: a for a in self.anchors}

    # Pen protocol support...

    def draw(self, pen):
        pen = PointToSegmentPen(pen)
        return self.drawPoints(pen)

    def drawPoints(self, pen):
        for path in self.paths:
            pen.beginPath()
            for node in path.nodes:
                pen.addPoint(
                    pt=(node.x, node.y),
                    segmentType=node.pen_type,
                    smooth=node.is_smooth,
                )
            pen.endPath()
        for component in self.components:
            pen.addComponent(component.ref, component.transform)

    def clearContours(self):
        self.shapes = []

    def getPen(self):
        return SegmentToPointPen(LayerPen(self))

    def decompose(self):
        pen = DecomposingRecordingPen(self._nested_component_dict())
        self.draw(pen)
        self.clearContours()
        pen.replay(self.getPen())


class LayerPen(AbstractPointPen):
    def __init__(self, target):
        self.target = target
        self.curPath = []

    def beginPath(self, identifier=None, **kwargs):
        self.curPath = []

    def endPath(self):
        """End the current sub path."""
        self.target.shapes.append(Shape(nodes=self.curPath))

    def addPoint(
        self, pt, segmentType=None, smooth=False, name=None, identifier=None, **kwargs
    ):
        if segmentType == "move":
            return
        ourtype = Node._from_pen_type[segmentType]
        if smooth:
            ourtype = ourtype + "s"
        self.curPath.append(Node(pt[0], pt[1], ourtype))

    def addComponent(self, baseGlyphName, transformation, identifier=None, **kwargs):
        self.target.shapes.append(Shape(ref=baseGlyphName, transform=transformation))



from babelfont.Anchor import Anchor
from babelfont.Axis import Axis
from babelfont.BaseObject import Color, I18NDictionary, OTValue, Position
from babelfont.convertors import BaseConvertor, Convert
from babelfont.Features import Features
from babelfont.Font import Font
from babelfont.Glyph import Glyph
from babelfont.Guide import Guide
from babelfont.Instance import Instance
from babelfont.Layer import Layer
from babelfont.Master import Master
from babelfont.Names import Names
from babelfont.Node import Node
from babelfont.Shape import Shape, Transform

__all__ = [
    "Font",
    "Axis",
    "Glyph",
    "Master",
    "Instance",
    "Guide",
    "Anchor",
    "Layer",
    "Shape",
    "Transform",
    "Node",
    "Names",
    "Color",
    "Position",
    "OTValue",
    "I18NDictionary",
    "Convert",
    "BaseConvertor",
    "Features",
    "load",
]


def load(filename):
    return Convert(filename).load()



import dataclasses
import re
import typing

from graphviz import Digraph

import babelfont
from babelfont import *
from babelfont.BaseObject import I18NDictionary

tocfile = open("docs/_data/navigation.yml", "w")
tocfile.write("default:\n")


def maybelink(t):
    if "babelfont" in str(t) and dataclasses.is_dataclass(t):
        return "[`%s`](%s.html)" % (t.__name__, t.__name__)
    if isinstance(t, typing._GenericAlias):
        if t._name == "Optional":
            return "Optional[%s]" % maybelink(t.__args__[0])
        elif t._name == "List":
            return "[%s]" % maybelink(t.__args__[0])
    if isinstance(t, tuple):
        return "(" + ", ".join([e.__name__ for e in t]) + ")"
    return t.__name__


def describe_dataclass(cls):
    if not dataclasses.is_dataclass(cls):
        return

    name = cls.__name__
    f = open("docs/%s.md" % name, "w")
    tocfile.write("  - title: %s\n    url: %s.html\n" % (name, name))
    f.write("---\ntitle: %s\n---\n" % name)

    f.write(cls.__doc__)
    f.write("\n")
    if cls._write_one_line:
        f.write(
            "* When writing to Babelfont-JSON, this class must be serialized without newlines\n"
        )
    for k in dataclasses.fields(cls):
        if k.name == "_":
            continue
        if isinstance(k.type, list):
            stringytype = "[%s]" % ", ".join([maybelink(t) for t in k.type])
        elif isinstance(k.type, tuple):
            stringytype = "(%s)" % ", ".join([maybelink(t) for t in k.type])
        else:
            stringytype = maybelink(k.type)
        if not "`" in stringytype:
            stringytype = "`%s`" % stringytype
        f.write("## %s.%s\n\n" % (name, k.name))
        f.write("* Python type: %s\n\n" % stringytype)
        if "json_type" in k.metadata:
            f.write("* Babelfont-JSON type: `%s`\n\n" % k.metadata["json_type"])
        if (
            k.default is dataclasses.MISSING
            and k.default_factory is dataclasses.MISSING
        ):
            f.write("* **Required field**\n\n")
        if "json_location" in k.metadata:
            f.write(
                "* When writing to Babelfont-JSON, this structure is stored under the separate file `%s`.\n\n"
                % k.metadata["json_location"]
            )

        if "separate_items" in k.metadata:
            f.write(
                "* When writing to Babelfont-JSON, each item in the list must be placed on a separate line.\n\n"
            )
        if "python_only" in k.metadata:
            f.write(
                "* This field only exists as an attribute of the the Python object and should not be written to Babelfont-JSON.\n\n"
            )

        if "description" in k.metadata:
            f.write(k.metadata["description"])
        if k.type == I18NDictionary:
            f.write(" *Localizable.*")
        f.write("\n")
        if k.default is not dataclasses.MISSING:
            f.write("*If not provided, defaults to* `%s`.\n" % str(k.default))
        f.write("\n\n")
    f.close()


describe_dataclass(Font)
describe_dataclass(Axis)
describe_dataclass(Instance)
describe_dataclass(Master)
describe_dataclass(Names)
describe_dataclass(Glyph)
describe_dataclass(Layer)
describe_dataclass(Guide)
describe_dataclass(Shape)
describe_dataclass(Anchor)


dot = Digraph(comment="Babelfont Format", format="svg")
dot.attr(rankdir="LR")
dot.attr(overlap="false")
donetypes = {}


def add_type_node(cls):
    def chktype(t, port):
        if "babelfont" in str(t) and dataclasses.is_dataclass(t):
            add_type_node(t)
            dot.edge(cls.__name__ + ":" + port, t.__name__)
        if isinstance(t, tuple):
            return "(" + ", ".join([e.__name__ for e in t]) + ")"
        return t.__name__

    if cls in donetypes:
        return
    if not dataclasses.is_dataclass(cls):
        return
    label = (
        """<<TABLE>
    <tr><td colspan="2" border="0"><b>%s</b></td></tr>
    """
        % cls.__name__
    )
    for k in dataclasses.fields(cls):
        if k.name == "_" or "python_only" in k.metadata:
            continue
        if isinstance(k.type, list):
            stringytype = "[%s]" % ", ".join([chktype(t, k.name) for t in k.type])
        elif isinstance(k.type, tuple):
            stringytype = "(%s)" % ", ".join([chktype(t, k.name) for t in k.type])
        else:
            stringytype = chktype(k.type, k.name)
        name = k.name
        if (
            k.default is dataclasses.MISSING
            and k.default_factory is dataclasses.MISSING
        ):
            name = "<b>%s</b>" % name
        label = label + '<tr><td>%s</td><td port="%s"><i>%s</i></td></tr>' % (
            name,
            k.name,
            stringytype,
        )

    label = label + "</TABLE>>"
    dot.node(
        cls.__name__,
        label,
        shape="none",
        fontname="Avenir",
        href="%s.html" % cls.__name__,
    )
    donetypes[cls] = True


add_type_node(Font)
add_type_node(Glyph)
dot.edge("Font:glyphs", "Glyph")
output = re.sub(r"(?s)^.*<svg", "<svg", dot.pipe().decode("utf-8"))
out = open("docs/index.md", "w")
out.write(
    """---
title: Babelfont Format
toc: false
---


"""
)
out.write(output + "\n")



import datetime
import math
import re
import uuid

import openstep_plist

from babelfont import (
    Anchor,
    Axis,
    Glyph,
    Guide,
    Instance,
    Layer,
    Master,
    Node,
    Shape,
    Transform,
)
from babelfont.BaseObject import OTValue
from babelfont.convertors import BaseConvertor
from babelfont.convertors.glyphs.utils import (
    _glyphs_metrics_to_ours,
    _stash,
)


class Glyphs2(BaseConvertor):
    suffix = ".glyphs"

    @classmethod
    def is_suitable_plist(cls, convertor):
        return (
            ".formatVersion" not in convertor.scratch["plist"]
            or convertor.scratch["plist"][".formatVersion"] < 3
        )

    @classmethod
    def can_save(cls, convertor, **kwargs):
        if not super().can_save(convertor, **kwargs):
            return False
        if "format" not in kwargs or kwargs["format"] != 2:
            return False
        return True

    @classmethod
    def can_load(cls, convertor):
        if not super().can_load(convertor):
            return False
        if "plist" not in convertor.scratch:
            convertor.scratch["plist"] = openstep_plist.load(
                open(convertor.filename, "r"), use_numbers=True
            )
        return cls.is_suitable_plist(convertor)

    def _load(self):
        self.glyphs = self.scratch["plist"]

        self.customParameters = {}
        for param in self.glyphs.get("customParameters", []):
            self.customParameters[param["name"]] = param["value"]

        self._load_axes()

        self._load_kern_groups(self.glyphs["glyphs"])

        for gmaster in self.glyphs["fontMaster"]:
            self.font.masters.append(self._load_master(gmaster))
        self._fixup_axes()
        if not self.font.default_master:
            raise ValueError("Cannot identify default master")

        for gglyph in self.glyphs["glyphs"]:
            g = self._load_glyph(gglyph)
            self.font.glyphs.append(g)

        for ginstance in self.glyphs.get("instances", []):
            self.font.instances.append(self._load_instance(ginstance))

        self._fixup_axis_mappings()

        self._load_metadata()
        self._load_features()
        return self.font

    def _load_axes(self):
        self.axis_name_map = {}
        if "Axes" in self.customParameters:
            for ax in self.customParameters["Axes"]:
                self.font.axes.append(Axis(name=ax["Name"], tag=ax["Tag"]))
                self.axis_name_map[ax["Name"]] = self.font.axes[-1]
        else:
            self.font.axes.append(Axis(name="Weight", tag="wght"))
            self.font.axes.append(Axis(name="Width", tag="wdth"))
            self.axis_name_map = {
                "Weight": self.font.axes[-1],
                "Width": self.font.axes[-1],
            }

    def _fixup_axes(self):
        for master in self.font.masters:
            for axis in self.font.axes:
                thisLoc = master.location[axis.tag]
                if axis.min is None or thisLoc < axis.min:
                    axis.min = thisLoc
                if master.id == self._default_master_id():
                    axis.default = master.location[axis.tag]
                if axis.max is None or thisLoc > axis.max:
                    axis.max = thisLoc

    def _fixup_axis_mappings(self):
        for axis in self.font.axes:
            if not axis.map:
                continue
            axis.map = list(sorted(set(axis.map)))
            axis.min, axis.max, axis.default = (
                axis.map_backward(axis.min),
                axis.map_backward(axis.max),
                axis.map_backward(axis.default),
            )

    def _custom_parameter(self, thing, name):
        for param in thing.get("customParameters", []):
            if param["name"] == name:
                return param["value"]
        return None

    def _default_master_id(self):
        # The default master in glyphs is either the first master or the
        # one selected by the Variable Font Origin custom parameter
        vfo = self._custom_parameter(self.glyphs, "Variable Font Origin")
        if vfo and vfo in self.font._master_map:
            return vfo
        return self.glyphs["fontMaster"][0]["id"]

    def _get_master_name(self, gmaster):
        if gmaster.get("name"):
            return gmaster["name"]
        cname = self._custom_parameter(gmaster, "Master Name")
        if cname:
            return cname
        # Remove None and empty string
        names = list(
            filter(
                None,
                [
                    gmaster.get("width", "Regular"),
                    gmaster.get("weight", "Regular"),
                    gmaster.get("custom", ""),
                ],
            )
        )

        # Remove redundant occurences of 'Regular'
        while len(names) > 1 and "Regular" in names:
            names.remove("Regular")
        if gmaster.get("italicAngle"):
            if names == ["Regular"]:
                return "Italic"
            if "Italic" not in gmaster.get("custom", ""):
                names.append("Italic")
        return " ".join(names)

    def _load_master(self, gmaster):
        # location = gmaster.get("axesValues", [])
        master = Master(
            name=self._get_master_name(gmaster),
            id=gmaster.get("id"),
        )
        for metric in Master.CORE_METRICS:
            master.metrics[metric] = gmaster.get(_glyphs_metrics_to_ours(metric))
        # Check for metrics in custom parameters

        master.font = self.font

        axisloc = self._custom_parameter(gmaster, "Axis Location")
        if axisloc:
            # I dunno, use that? Needs mapping? Check we are using tags/IDs
            location = axisloc
        else:
            potential_locations = [
                gmaster.get("weightValue", 100),
                gmaster.get("widthValue", 100),
                gmaster.get("customValue", 0),
                gmaster.get("customValue1", 0),
                gmaster.get("customValue2", 0),
                gmaster.get("customValue3", 0),
            ]
            location = {}
            for k, loc in zip(self.font.axes, potential_locations):
                location[k.tag] = loc
        master.location = location

        master.guides = [self._load_guide(x) for x in gmaster.get("guides", [])]

        kernmaster = None
        if self._custom_parameter(gmaster, "Link Metrics With First Master"):
            kernmaster = self.glyphs["fontMaster"][0]["id"]
        elif self._custom_parameter(gmaster, "Link Metrics With Master"):
            kernmaster_name = self._custom_parameter(
                gmaster, "Link Metrics With Master"
            )
            for m in self.glyphs["fontMaster"]:
                name = self._get_master_name(m)
                if name == kernmaster_name:
                    kernmaster = m["id"]
        else:
            kernmaster = master.id
        kerntable = self.glyphs.get("kerning", {}).get(kernmaster, {})
        master.kerning = self._load_kerning(kerntable)

        _stash(master, gmaster)
        return master

    def _get_codepoint(self, gglyph):
        cp = gglyph.get("unicode")
        if not cp:
            return []
        if isinstance(cp, int):
            return [int("%04i" % cp, 16)]
        return [int(x, 16) for x in cp.split(",")]

    def _load_glyph(self, gglyph):
        name = gglyph["glyphname"]
        c = gglyph.get("category")
        sc = gglyph.get("subCategory")
        if sc == "Ligature":
            category = "ligature"
        if c == "Mark":
            category = "mark"
        else:
            category = "base"
        cp = self._get_codepoint(gglyph)
        exported = True
        if "export" in gglyph and gglyph["export"] == 0:
            exported = False
        g = Glyph(name=name, codepoints=cp or [], category=category, exported=exported)
        g.production_name = gglyph.pop("production", None)

        for layer in gglyph.pop("layers"):
            g.layers.extend(self._load_layer(layer, g))

        _stash(g, gglyph)
        return g

    def _load_layer(self, layer, glyph, width=None):
        if width is None:
            width = layer["width"]
        l = Layer(width=width, id=layer.get("layerId"), _font=self.font, _glyph=glyph)
        l.name = layer.get("name")
        if [x for x in self.font.masters if x.id == l.id]:
            l._master = l.id
        else:
            l._master = layer.get("associatedMasterId")
        l.guides = [
            self._load_guide(x)
            for x in layer.get("guideLines", layer.get("guides", []))
        ]
        l.shapes = []
        for shape in layer.get("shapes", []):
            l.shapes.append(self._load_shape(shape))
        for shape in layer.get("paths", []):
            l.shapes.append(self._load_path(shape))
        for shape in layer.get("components", []):
            l.shapes.append(self._load_component(shape))
        for anchor in layer.get("anchors", []):
            l.anchors.append(self._load_anchor(anchor))

        returns = [l]
        if "background" in layer:
            (background,) = self._load_layer(layer["background"], width=l.width)
            # If it doesn't have an ID, we need to generate one
            background.id = background.id or str(uuid.uuid1())
            background.isBackground = True

            l.background = background.id
            returns.append(background)
        # TODO backgroundImage, metricTop/Bottom/etc, vertOrigin, vertWidth.
        for r in returns:
            assert r.valid
        _stash(l, layer)

        return returns

    def _load_guide(self, gguide):
        pos = gguide.get("position", "{0, 0}")
        m = re.match(r"^\{(\S+), (\S+)\}", pos)
        return Guide(pos=[int(m[1]), int(m[2]), int(gguide.get("angle", 0))])

    def _load_anchor(self, ganchor):
        pos = ganchor.get("position", "{0, 0}")
        m = re.match(r"^\{(\S+), (\S+)\}", pos)
        return Anchor(name=ganchor["name"], x=float(m[1]), y=float(m[2]))

    def _load_instance(self, ginstance):
        if ginstance.get("type") == "variable":
            return
        if "axesValues" in ginstance:
            location = ginstance["axesValues"]
            instance_location = {k.tag: v for k, v in zip(self.font.axes, location)}
        elif "instanceInterpolations" in ginstance:
            # All right then.
            instance_location = {k.tag: 0 for k in self.font.axes}
            for mId, factor in ginstance["instanceInterpolations"].items():
                master_loc = self.font.master(mId).location
                for k in self.font.axes:
                    instance_location[k.tag] += master_loc[k.tag] * factor
        else:
            raise ValueError("Need to Synthesize location")
        i = Instance(
            name=ginstance["name"],
            styleName=ginstance["name"],
            location=instance_location,
        )
        c = self._custom_parameter(ginstance, "Axis Location") or []
        for loc in c:
            ax = self.axis_name_map[loc["Axis"]]
            if not ax.map:
                ax.map = []
            ax.map.append(
                (
                    int(loc["Location"]),
                    instance_location[ax.tag],
                )
            )

        _maybesetformatspecific(i, ginstance, "customParameters")
        return i

    def _load_path(self, path):
        shape = Shape()
        shape.nodes = []
        for n in path["nodes"]:
            m = re.match(r"(\S+) (\S+) (\S+)( SMOOTH)?(.*)", n)
            ntype = m[3][0].lower()
            if m[4]:
                ntype = ntype + "s"
            n = Node(x=float(m[1]), y=float(m[2]), type=ntype)
            shape.nodes.append(n)
        shape.closed = bool(path["closed"])
        _maybesetformatspecific(shape, path, "attr")
        return shape

    def _load_component(self, shape):
        glyphname = shape.get("ref", shape.get("name"))
        transform = shape.get("transform")
        if isinstance(transform, str):
            m = re.match(r"^\{(\S+), (\S+), (\S+), (\S+), (\S+), (\S+)\}", transform)
            transform = Transform(*[float(g) for g in m.groups()])
        c = Shape(ref=glyphname)

        if not transform:
            translate = Transform().translate(*shape.get("pos", (0, 0)))
            scale = Transform().scale(*shape.get("scale", (1, 1)))
            rotation = Transform().rotate(math.radians(shape.get("angle", 0)))
            # Compute transform...
            transform = translate.transform(scale).transform(rotation)

        c.transform = transform
        for entry in [
            "alignment",
            "anchorTo",
            "attr",
            "locked",
            "orientation",
            "piece",
            "userData",
        ]:
            _maybesetformatspecific(c, shape, entry)
        return c

    def _load_kern_groups(self, glyphs):
        kerngroups = {}
        for g in glyphs:
            l_class = g.get("leftKerningGroup", g["glyphname"])
            r_class = g.get("rightKerningGroup", g["glyphname"])
            # DAMMIT GLYPHS
            kerngroups.setdefault("MMK_R_" + l_class, []).append(g["glyphname"])
            kerngroups.setdefault("MMK_L_" + r_class, []).append(g["glyphname"])
        for k, v in kerngroups.items():
            self.font.features.classes[k] = tuple(v)

    def _load_kerning(self, kerndict):
        return {
            (l, r): value
            for l, level2 in kerndict.items()
            for r, value in level2.items()
        }

    def _load_metadata(self):
        self.font.upm = self.glyphs["unitsPerEm"]
        self.font.version = (self.glyphs["versionMajor"], self.glyphs["versionMinor"])
        self.font.names.familyName.set_default(self.glyphs["familyName"])

        # This is very glyphs 3
        props = {}
        for prop in self.glyphs.get("properties", []):
            if "value" in prop:
                props[prop["key"]] = prop["value"]
            else:
                props[prop["key"]] = {p["language"]: p["value"] for p in prop["values"]}

        if props:
            interestingProps = {
                "copyrights": "copyright",
                "designer": "designer",
                "designerURL": "designerURL",
            }  # Etc
            for glyphsname, attrname in interestingProps.items():
                thing = props.get(glyphsname, "")
                if isinstance(thing, dict):
                    getattr(self.font.names, attrname).copy_in(thing)
                else:
                    getattr(self.font.names, attrname).set_default(thing)
            # Do other properties here

        # Any customparameters in the default master which look like
        # custom OT values need to move there.
        # cp = self.font.default_master._formatspecific.get("com.glyphsapp", {}).get(
        #     "customParameters", {}
        # )
        # for param in cp:
        #     ot_param = opentype_custom_parameters.get(param["name"])
        #     if not ot_param:
        #         continue
        #     self.font.customOpenTypeValues.append(
        #         OTValue(ot_param[0], ot_param[1], param["value"])
        #     )

        self.font.note = self.glyphs.get("note")
        self.font.date = datetime.datetime.strptime(
            self.glyphs.get("date"), "%Y-%m-%d %H:%M:%S +0000"
        )
        _stash(self.font, self.glyphs)

    def _load_features(self):
        for c in self.glyphs.get("classes", []):
            self.font.features.classes[c["name"]] = tuple(c["code"].split())

        # featurefile = ""
        # for f in self.glyphs.get("featurePrefixes", []):
        #     featurefile = featurefile + f.get("code")
        # for f in self.glyphs.get("features", []):
        #     tag = f.get("tag", f.get("name", ""))
        #     feacode = "feature %s { %s\n} %s;" % (tag, f["code"], tag)
        #     featurefile = featurefile + feacode

        # glyphNames = {g.name for g in self.font.glyphs}
        # feaparser = FeaParser(featurefile, glyphNames=glyphNames)
        # ast = feaparser.parser.ast
        # for name, members in self.font.features.classes.items():
        #     glyphclass = ast.GlyphClassDefinition(
        #         name, ast.GlyphClass([m for m in members])
        #     )
        #     feaparser.parser.glyphclasses_.define(name, glyphclass)
        # feaparser.parse()
        # self.font.features += feaparser.ff



# encoding: utf-8
from __future__ import division, print_function, unicode_literals
import objc
from GlyphsApp import *
from GlyphsApp.plugins import *
import sys
import os

from babelfont.convertors.gsobject import GSObject
from babelfont import Font

class BabelfontExport(FileFormatPlugin):
    # The NSView object from the User Interface. Keep this here!
    dialog = objc.IBOutlet()

    @objc.python_method
    def settings(self):
        self.name = "Babelfont"
        self.icon = "ExportIconTemplate"
        self.toolbarPosition = 100
        # Load .nib dialog (with .extension)
        # self.loadNib("IBdialog", __file__)

    # @objc.python_method
    # def start(self):
    #   pass

    @objc.python_method
    def export(self, font):
        # Ask for export destination and write the file:
        title = "Choose export destination"
        proposedFilename = font.familyName
        fileTypes = ["babelfont"]
        # Call dialog
        filepath = GetSaveFile(title, proposedFilename, fileTypes)

        if filepath:
            f = GSObject()
            f.scratch = {"gsfont": font}
            f.font = Font()
            f.gsfont = font
            f._load()
            f.font.save(filepath)
            return (
                True,
                'The export of "%s" was successful.' % (os.path.basename(filepath)),
            )

        else:
            return (False, "No file chosen")

    def __file__(self):
        """Please leave this method unchanged"""
        return __file__



import functools
import logging
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List

from fontTools.feaLib.variableScalar import VariableScalar
from fontTools.varLib.models import VariationModel

from .Axis import Axis
from .BaseObject import BaseObject, IncompatibleMastersError, OTValue
from .Features import Features
from .Glyph import GlyphList
from .Instance import Instance
from .Master import Master
from .Names import Names

log = logging.getLogger(__name__)


@dataclass
class _FontFields:
    upm: int = field(default=1000, metadata={"description": "The font's units per em."})
    version: (int, int) = field(
        default=(1, 0),
        metadata={
            "description": "Font version number as a tuple of integers (major, minor).",
            "json_type": "[int,int]",
        },
    )
    axes: [Axis] = field(
        default_factory=list,
        metadata={
            "separate_items": True,
            "description": "A list of axes, in the case of variable/multiple master font. May be empty.",
        },
    )
    instances: [Instance] = field(
        default_factory=list,
        metadata={
            "separate_items": True,
            "description": "A list of named/static instances.",
        },
    )
    masters: [Master] = field(
        default_factory=list,
        metadata={
            "separate_items": True,
            "description": "A list of the font's masters.",
        },
    )
    glyphs: GlyphList = field(
        default_factory=GlyphList,
        metadata={
            "skip_serialize": True,
            "separate_items": True,
            "json_type": "[dict]",
            "json_location": "glyphs.json",
            "description": """A list of all glyphs supported in the font.

The `GlyphList` structure in the Python object is a dictionary with array-like
properties (or you might think of it as an array with dictionary-like properties)
containing [`Glyph`](Glyph.html) objects. The `GlyphList` may be iterated
directly, and may be appended to, but may also be used to index a `Glyph` by
its name. This is generally what you want:

```Python

for g in font.glyphs:
    assert isinstance(g, Glyph)

font.glyphs.append(newglyph)

glyph_ampersand = font.glyphs["ampersand"]
```
            """,
        },
    )
    note: str = field(
        default=None,
        metadata={"description": "Any user-defined textual note about this font."},
    )
    date: datetime = field(
        default_factory=datetime.now,
        metadata={
            "description": """The font's date. When writing to Babelfont-JSON, this
should be stored in the format `%Y-%m-%d %H:%M:%S`. *If not provided, defaults
to the current date/time*.""",
            "json_type": "str",
        },
    )
    names: Names = field(default_factory=Names, metadata={"skip_serialize": True})
    customOpenTypeValues: [OTValue] = field(
        default_factory=list,
        metadata={
            "description": "Any values to be placed in OpenType tables on export to override defaults; these must be font-wide. Metrics which may vary by master should be placed in the `metrics` field of a Master."
        },
    )
    features: Features = field(
        default_factory=Features,
        metadata={
            "description": "A representation of the font's OpenType features",
        },
    )
    first_kern_groups: Dict[str, List[str]] = field(
        default_factory=dict,
        metadata={
            "description": "A dictionary of kerning groups, where the key is the group name and the value is a list of glyph names in the group."
        },
    )
    second_kern_groups: Dict[str, List[str]] = field(
        default_factory=dict,
        metadata={
            "description": "A dictionary of kerning groups, where the key is the group name and the value is a list of glyph names in the group."
        },
    )


@dataclass
class Font(_FontFields, BaseObject):
    """Represents a font, with one or more masters."""

    def __repr__(self):
        return "<Font '%s' (%i masters)>" % (
            self.names.familyName.get_default(),
            len(self.masters),
        )

    def save(self, filename, **kwargs):
        from .convertors import Convert

        return Convert(filename).save(self, **kwargs)

    def master(self, mid):
        return self._master_map[mid]

    def map_forward(self, location):
        location2 = dict(location)
        for a in self.axes:
            if a.tag in location2:
                location2[a.tag] = a.map_forward(location2[a.tag])
        return location2

    def map_backward(self, location):
        location2 = dict(location)
        for a in self.axes:
            if a.tag in location2:
                location2[a.tag] = a.map_backward(location2[a.tag])
        return location2

    @functools.cached_property
    def default_master(self):
        default_loc = {a.tag: a.userspace_to_designspace(a.default) for a in self.axes}
        for m in self.masters:
            if m.location == default_loc:
                return m
        if len(self.masters) == 1:
            return self.masters[0]
        raise ValueError("Could not determine default master")

    @functools.cached_property
    def _master_map(self):
        return {m.id: m for m in self.masters}

    @functools.cached_property
    def unicode_map(self):
        unicodes = {}
        for g in self.glyphs:
            if not g.codepoints:
                continue
            for u in g.codepoints:
                if u:
                    unicodes[u] = g.name
        return unicodes

    def variation_model(self):
        return VariationModel(
            [m.normalized_location for m in self.masters],
            axisOrder=[a.tag for a in self.axes],
        )

    @functools.cached_property
    def _all_kerning(self):
        all_keys = [set(m.kerning.keys()) for m in self.masters]
        kerndict = {}
        for left, right in list(set().union(*all_keys)):
            kern = VariableScalar()
            kern.axes = self.axes
            for m in self.masters:
                thiskern = m.kerning.get((left, right), 0)
                if (left, right) not in m.kerning:
                    log.debug(
                        "Master %s did not define a kern pair for (%s, %s), using 0"
                        % (m.name.get_default(), left, right)
                    )
                kern.add_value(m.location, thiskern)
            kerndict[(left, right)] = kern
        return kerndict

    @functools.cached_property
    def _all_anchors(self):
        _all_anchors_dict = {}
        for g in sorted(self.glyphs.keys()):
            default_layer = self.default_master.get_glyph_layer(g)
            has_mark = None
            for a in default_layer.anchors_dict.keys():
                if a[0] == "_":
                    if has_mark:
                        log.warning(
                            "Glyph %s tried to be in two mark classes (%s, %s). The first one will win."
                            % (g, has_mark, a)
                        )
                        continue
                    has_mark = a
                if a not in _all_anchors_dict:
                    _all_anchors_dict[a] = {}
                _all_anchors_dict[a][g] = self.get_variable_anchor(g, a)
        return _all_anchors_dict

    def get_variable_anchor(self, glyph, anchorname):
        x_vs = VariableScalar()
        x_vs.axes = self.axes
        y_vs = VariableScalar()
        y_vs.axes = self.axes
        for ix, m in enumerate(self.masters):
            layer = m.get_glyph_layer(glyph)
            if anchorname not in layer.anchors_dict:
                raise IncompatibleMastersError(
                    "Anchor %s not found on glyph %s in master %s"
                    % (anchorname, glyph, m)
                )
            anchor = m.get_glyph_layer(glyph).anchors_dict[anchorname]
            x_vs.add_value(self.map_forward(m.location), anchor.x)
            y_vs.add_value(self.map_forward(m.location), anchor.y)
        return (x_vs, y_vs)

    def exportedGlyphs(self):
        return [g.name for g in self.glyphs if g.exported]



from babelfont.convertors.glyphs.glyphs2 import Glyphs2
from babelfont.convertors.glyphs.glyphs3 import GlyphsThree
from babelfont.convertors.glyphs.glyphspackage import GlyphsPackage

__all__ = [
    "Glyphs2",
    "GlyphsThree",
    "GlyphsPackage",
]



import datetime
import re
import math

import orjson
from fontTools.misc.transform import Transform

from babelfont import Axis, Glyph, Layer, Master, Node, Shape, Guide
from babelfont.BaseObject import Color, I18NDictionary, OTValue
from babelfont.convertors import BaseConvertor


class FontlabVFJ(BaseConvertor):
    suffix = ".vfj"

    def _load(self):
        self.fontlab = orjson.loads(open(self.filename).read())["font"]
        self.known_transforms = {}
        self._load_axes()
        self._load_masters()
        if "defaultMaster" in self.fontlab:
            default = self.font.master(self.fontlab["defaultMaster"])
            if self.font.axes:
                def_loc = default.location
                for axis in self.font.axes:
                    if axis.tag in default.location:
                        axis.default = axis.designspace_to_userspace(def_loc[axis.tag])
            assert self.font.default_master
        for g in self.fontlab.get("glyphs", []):
            glyph = self._load_thing(g, self.glyph_loader)
            # Fontlab allows glyphs to elide layers with no shapes
            # so we need to ensure there is a layer for each master
            layer_ids = [l.id for l in glyph.layers]
            for master in self.font.masters:
                if master.id not in layer_ids:
                    layer = Layer(
                        name=master.id,
                        _master=master.id,
                    )
                    layer._font = self.font
                    glyph.layers.append(layer)

            self.font.glyphs.append(glyph)
        for cls in self.fontlab.get("classes", []):
            self._load_class(cls)
        for feature in self.fontlab.get("openTypeFeatures", []):
            self._load_feature(feature)
        self.font.upm = self.fontlab.get("upm", 1000)
        self._load_metadata()
        return self.font

    def _load_feature(self, feature):
        count = 0
        if "name" in feature:
            # Remove the "feature ... {" first line and "} tag;" last line
            feature["feature"] = re.sub(
                r"^\s*feature\s+(\w+)\s*{(.*)\s*}\s*\1\s*;",
                r"\2",
                feature["feature"],
                flags=re.DOTALL,
            )
            self.font.features.features.append((feature["name"], feature["feature"]))
        else:
            self.font.features.prefixes["Anonymous prefix " + str(count)] = feature[
                "feature"
            ]
            count += 1

    def _load_class(self, cls):
        self.font.features.classes[cls["name"]] = cls["names"]

    axis_loader = (
        Axis,
        {
            "name": "name",
            "tag": "tag",
            "minimum": "min",
            "maximum": "max",
            "default": "default",
            "axisGraph": ("map", lambda _, x: [(v, int(k)) for k, v in x.items()]),
        },
    )

    def _load_kerning(self, flkerning):
        for kclass in flkerning.get("kerningClasses", []):
            if kclass.get("1st"):
                self.font.first_kern_groups[kclass["name"]] = kclass["names"]
            if kclass.get("2nd"):
                self.font.second_kern_groups[kclass["name"]] = kclass["names"]
        kerning = {}
        for left, rvalue in flkerning.get("pairs", {}).items():
            if left.startswith("@"):
                left = "@first_group_" + left[1:]
            for right, value in rvalue.items():
                if right.startswith("@"):
                    right = "@second_group_" + right[1:]
                kerning[(left, right)] = int(value)
        return kerning

    def _convert_color(_, col):
        r, g, b = int(col[1:3], 16), int(col[3:5], 16), int(col[5:7], 16)
        return Color(r, g, b, 0)

    def _convert_shapes(self, flshapes):
        shapes = []
        for shape in flshapes:
            transform_j = shape.get("transform", {})
            if "id" in transform_j:
                self.known_transforms[transform_j["id"]] = transform_j
            if isinstance(transform_j, str):
                transform_j = self.known_transforms[transform_j]
            transform = Transform().translate(
                transform_j.get("xOffset", 0),
                transform_j.get("yOffset", 0),
            )
            if "component" in shape:
                shapes.append(
                    Shape(
                        ref=shape["component"]["glyphName"],
                        transform=transform,
                        _=shape["component"],
                    )
                )
            else:
                for flcontour in shape["elementData"].get("contours", []):
                    contour = Shape(nodes=[])
                    for n in flcontour["nodes"]:
                        contour.nodes.extend(self._load_node(n))
                    if (
                        contour.nodes[0].type[0] == "l"
                        and contour.nodes[-1].type[0] == "q"
                    ):
                        contour.nodes[0].type = "cs"
                        contour.nodes[-1].type = "o"
                    shapes.append(contour)
        return shapes

    def _convert_guides(self, flguides):
        guides = []
        for guide in flguides:
            if "horizontal" in guide:
                y = guide["position"]
                x = 0
                angle = 0
            elif "vertical" in guide:
                x = guide.get("position", 0)
                y = 0
                angle = 90
            else:
                x, y = [float(v) for v in guide["center"].split(" ")]
                vector_x, vector_y = [float(v) for v in guide["vector"].split(" ")]
                angle = math.degrees(math.atan2(vector_y - y, vector_x - x))
            guides.append(Guide(pos=(x, y, angle)))
        return guides

    def _load_node(self, input_):
        stuff = input_.split("  ")
        rv = []
        for ix, node in enumerate(stuff):
            m = re.match(r"(-?[\d\.]+) (-?[\d\.]+)( s)?", node)
            if not m:
                raise ValueError("Can't understand nodestring %s" % node)
            if len(stuff) == 1:
                ntyp = "l"
            elif len(stuff) == 3 and ix == 2:
                ntyp = "c"
            elif len(stuff) == 2 and ix == 1:
                ntyp = "q"
            else:
                ntyp = "o"
            if m[3]:
                ntyp = ntyp + "s"
            rv.append(Node(x=float(m[1]), y=float(m[2]), type=ntyp))
        return rv

    # The way these loaders work is they define a class to load into and a schema.
    # The schema is a mapping between *Fontlab* field names, and what we do with
    # it: if it's a plain string, this is our field name; if it's a tuple, it's
    # our field name and a method to transform the value; if it's a list, it's
    # a combination of field/tuples.
    master_loader = (
        Master,
        {
            "name": ["id", "name"],
            "kerning": ("kerning", _load_kerning),
            "location": (
                "location",
                lambda s, x: {s.axis_name_map[k].tag: v for k, v in x.items()},
            ),
        },
    )

    _layer_loader = (
        Layer,
        {
            "advanceWidth": "width",
            "name": ["name", "_master", "id"],
            "color": ("color", _convert_color),
            "elements": ("shapes", _convert_shapes),
            "guidelines": ("guides", _convert_guides),
        },
    )

    glyph_loader = (
        Glyph,
        {
            "name": "name",
            "unicode": (
                "codepoints",
                lambda _, x: [int(cp, 16) for cp in x.split(",")],
            ),
            "layers": ("layers", _layer_loader),
            "elements": None,
            "openTypeGlyphClass": (
                "category",
                lambda _, x: [None, "base", "ligature", "mark", "component"][x],
            ),
        },
    )

    def _load_thing(self, thing, handler):
        kwargs = {}
        kwargs["_"] = {}
        cls, mapping = handler
        for k, v in thing.items():
            if k not in mapping:
                kwargs["_"][k] = v
            elif mapping[k] is None:
                continue
            elif isinstance(mapping[k], list):
                for newname in mapping[k]:
                    kwargs[newname] = v
            elif isinstance(mapping[k], str):
                kwargs[mapping[k]] = v
            elif isinstance(mapping[k], tuple) and callable(mapping[k][1]):
                newName, convertor = mapping[k]
                kwargs[newName] = convertor(self, v)
            else:
                newName, convertor = mapping[k]
                if isinstance(v, list):
                    kwargs[newName] = [self._load_thing(v1, convertor) for v1 in v]
                else:
                    kwargs[newName] = self._load_thing(v, convertor)
        obj = cls(**kwargs)
        obj._font = self.font
        return obj

    def _load_axes(self):
        self.axis_name_map = {}
        for ax in self.fontlab.get("axes", []):
            axis = self._load_thing(ax, self.axis_loader)
            self.axis_name_map[ax["shortName"]] = axis
            self.font.axes.append(axis)

    def _load_masters(self):
        for m in self.fontlab.get("masters", []):
            fl_master = m["fontMaster"]
            metric_dict = {}
            for metric in [
                "ascender",
                "descender",
                "xHeight",
                "capsHeight",
                "measurement",
                "linegap",
                "underlineThickness",
                "underlinePosition",
            ]:
                if metric not in fl_master:
                    continue
                value = fl_master[metric]
                del fl_master[metric]
                if metric == "capsHeight":
                    metric = "capHeight"
                if metric == "linegap":
                    metric = "hheaLineGap"
                metric_dict[metric] = value
            master = self._load_thing(fl_master, self.master_loader)
            master.metrics = metric_dict
            master.font = self.font
            master.guides = self._convert_guides(fl_master.get("guidelines", []))
            self.font.masters.append(master)

    def _load_metadata(self):
        info = self.fontlab.get("info", {})
        self.font.names.familyName = I18NDictionary.with_default(info.get("tfn"))
        # sgn?
        for tag in [
            "copyright",
            "designer",
            "designerURL",
            "license",
            "licenseURL",
            "manufacturer",
            "manufacturerURL",
            "version",
        ]:
            setattr(self.font.names, tag, I18NDictionary.with_default(info.get(tag)))
        self.font.version = (info.get("versionMajor"), info.get("versionMinor"))
        if "vendor" in info:
            self.font.customOpenTypeValues.append(
                OTValue("OS/2", "achVendID", info["vendor"])
            )
        # Fontlab date format = "2020/08/28 15:33:36"
        if "date" in info:
            self.font.date = datetime.datetime.strptime(
                info["creationDate"],
                "%Y/%m/%d %H:%M:%S",
            )
        if "useTypoMetrics" in info:
            self.font.customOpenTypeValues.append(OTValue("OS/2", "fsSelection", 0x7))
        if "embedding" in info:
            self.font.customOpenTypeValues.append(
                OTValue("OS/2", "fsType", int(info["embedding"], 16))
            )
        # codepageRange
        # unicodeRange



from babelfont.Font import Font

def zero_background_width(font: Font):
    for glyph in font.glyphs:
        for layer in glyph.layers:
            if layer.isBackground:
                layer.width = 0

def decompose_backgrounds(font: Font):
    for glyph in font.glyphs:
        for layer in glyph.layers:
            if layer.isBackground:
                layer.decompose()